"""Configuration parameters for psy-view

**Disclaimer**

Copyright (C) 2020 Philipp S. Sommer

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see https://www.gnu.org/licenses/.
"""
from __future__ import annotations
from typing import (
    Dict,
    List,
    Any,
    Optional,
)

from psyplot_gui.config.rcsetup import (
    RcParams, validate_stringlist, psyplot_fname)
from psyplot.config.rcsetup import validate_dict


defaultParams: Dict[str, List[Any]] = {
    "projections": [
        ["cf", "cyl", "robin", "ortho", "moll"], validate_stringlist,
        "The names of available projections"],
    "savefig_kws": [
        dict(dpi=250), validate_dict,
        "Options that are passed to plt.savefig when exporting images"],
    "animations.export_kws": [
        dict(writer="ffmpeg"), validate_dict,
        "Options that are passed to FuncAnimation.save"],
    }


class PsyViewRcParams(RcParams):
    """RcParams for the psyplot-gui package."""

    HEADER: str = RcParams.HEADER.replace(
        'psyplotrc.yml', 'psyviewrc.yml').replace(
            'PSYVIEWRC', 'psyviewrc.yml')

    def load_from_file(self, fname: Optional[str] = None):
        """
        Update rcParams from user-defined settings

        This function updates the instance with what is found in `fname`

        Parameters
        ----------
        fname: str
            Path to the yaml configuration file. Possible keys of the
            dictionary are defined by :data:`config.rcsetup.defaultParams`.
            If None, the :func:`config.rcsetup.psyplot_fname` function is used.

        See Also
        --------
        dump_to_file, psyplot_fname"""
        fname = fname or psyplot_fname(env_key='PSYVIEWRC',
                                       fname='psyviewrc.yml')
        if fname:
            super().load_from_file(fname)


rcParams = PsyViewRcParams(defaultParams=defaultParams)
rcParams.update_from_defaultParams()
rcParams.load_from_file()



"""Dialogs for manipulating formatoptions.

**Disclaimer**

Copyright (C) 2020 Philipp S. Sommer

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see https://www.gnu.org/licenses/.
"""
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Dict,
    Any,
    Optional,
    Tuple,
    Union,
    Type,
)

import yaml
from PyQt5 import QtWidgets, QtGui
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas)
from matplotlib.figure import Figure

from psyplot.plotter import Plotter, Formatoption

if TYPE_CHECKING:
    from psyplot.project import Project
    from PyQt5.QtCore import QEvent  # pylint: disable=no-name-in-module


#: TODO: Find a more appropriate description here
Color = Any

#: TODO: Find a more appropriate description here
LSM_T = Dict[str, Any]


class BasemapDialog(QtWidgets.QDialog):
    """A dialog to modify the basemap settings."""

    xgrid_value: Optional[Union[str, Tuple[Any, Any]]]
    ygrid_value: Optional[Union[str, Tuple[Any, Any]]]

    def __init__(self, plotter: Plotter, *args, **kwargs) -> None:
        """
        Parameters
        ----------
        plotter: psy_maps.plotters.MapPlotter
            The psyplot plotter to configure
        """
        import psy_simple.widgets.colors as pswc
        import pandas as pd
        super().__init__(*args, **kwargs)
        vbox = QtWidgets.QVBoxLayout(self)

        #: colors that affect the map background
        self.colors = ['background', 'land', 'ocean', 'coast']

        #: QGridLayout to display the various colors
        grid = QtWidgets.QGridLayout()

        defaults = self.default_colors

        #: :class:`pandas.DataFrame` of widgets to modifiy the :attr:`colors`
        self.widgets = widgets = pd.DataFrame(
            index=['enable', 'color'], columns=self.colors, dtype=object)

        for i, col in enumerate(self.colors):
            widgets.iloc[0, i] = cb = QtWidgets.QCheckBox()
            cb.setChecked(False)
            widgets.iloc[1, i] = lbl = pswc.ColorLabel(defaults[col])
            lbl.setEnabled(False)

            cb.stateChanged.connect(lbl.setEnabled)

            grid.addWidget(QtWidgets.QLabel(col), 0, i)
            grid.addWidget(cb, 1, i)
            grid.addWidget(lbl, 2, i)

        vbox.addLayout(grid)

        #: Button box to cancel the operator or update the plotter
        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            self)

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        proj_box = QtWidgets.QGroupBox("Projection settings")
        layout = QtWidgets.QFormLayout(proj_box)

        #: text box for the central longitude (clon formatoption)
        self.txt_clon = QtWidgets.QLineEdit()
        self.txt_clon.setPlaceholderText('auto')
        self.txt_clon.setToolTip('Central longitude in degrees East')
        self.txt_clon.setValidator(QtGui.QDoubleValidator(-360, 360, 7))
        layout.addRow('Central longitude: ', self.txt_clon)

        #: text box for the central latitude (clat formatoption)
        self.txt_clat = QtWidgets.QLineEdit()
        self.txt_clat.setPlaceholderText('auto')
        self.txt_clat.setToolTip('Central latitude in degrees North')
        self.txt_clat.setValidator(QtGui.QDoubleValidator(-90, 90, 7))
        layout.addRow('Central latitude: ', self.txt_clat)

        vbox.addWidget(proj_box)

        #: group box for modifying the resolution of the land-sea-mask, see
        #: :attr:`opt_110m`, :attr:`opt_50m`, :attr:`opt_10m`
        self.lsm_box = QtWidgets.QGroupBox('Coastlines')
        self.lsm_box.setCheckable(True)
        hbox = QtWidgets.QHBoxLayout(self.lsm_box)
        hbox.addWidget(QtWidgets.QLabel("Resolution:"))

        #: Radiobutton for 110m resolution of lsm
        self.opt_110m = QtWidgets.QRadioButton("110m")

        #: Radiobutton for 50m resolution of lsm
        self.opt_50m = QtWidgets.QRadioButton("50m")

        #: Radiobutton for 10m resolution of lsm
        self.opt_10m = QtWidgets.QRadioButton("10m")
        hbox.addWidget(self.opt_110m)
        hbox.addWidget(self.opt_50m)
        hbox.addWidget(self.opt_10m)

        vbox.addWidget(self.lsm_box)

        #: group box drawing grid lines and labels
        self.grid_labels_box = QtWidgets.QGroupBox('Labels')
        self.grid_labels_box.setToolTip("Draw labels of meridionals and "
                                        "parallels")
        self.grid_labels_box.setCheckable(True)

        #: text box for the fontsize of grid labels
        self.txt_grid_fontsize = QtWidgets.QLineEdit()

        form = QtWidgets.QFormLayout(self.grid_labels_box)
        form.addRow("Font size:", self.txt_grid_fontsize)

        vbox.addWidget(self.grid_labels_box)

        #: Group box for options specific to meridionals (see
        #: :attr:`opt_meri_auto`, :attr:`opt_meri_at` and
        #: :attr:`opt_meri_every`, :attr:`opt_meri_num`)
        self.meridionals_box = QtWidgets.QGroupBox('Meridionals')
        self.meridionals_box.setCheckable(True)

        #: Radiobutton for automatic drawing of meridionals
        self.opt_meri_auto = QtWidgets.QRadioButton("auto")

        #: Radiobutton for giving the exact position of meridionals (see
        #: :attr:`txt_meri_at`)
        self.opt_meri_at = QtWidgets.QRadioButton("At:")

        #: Text field to enter the location of the meridionals on the map (see
        #: :attr:`opt_meri_at`)
        self.txt_meri_at = QtWidgets.QLineEdit()
        self.txt_meri_at.setPlaceholderText("30, 60, 90, 120, ... °E")
        # TODO: Add validator

        #: Radiobutton for equal-width spaced meridionals (see
        #: :attr:`txt_meri_every`)
        self.opt_meri_every = QtWidgets.QRadioButton("Every:")

        #: Text box to specify the distance between two meridionals (see
        #: :attr:`opt_meri_every`)
        self.txt_meri_every = QtWidgets.QLineEdit()
        self.txt_meri_every.setPlaceholderText("30 °E")
        self.txt_meri_every.setValidator(QtGui.QDoubleValidator(0, 360, 7))

        #: Radiobutton to draw a specific number of meridionals with
        #: equal-distance (see also :attr:`txt_meri_num`)
        self.opt_meri_num = QtWidgets.QRadioButton("Number:")

        #: Text box to set the number of meridionals to be shown (see
        #: :attr:`opt_meri_num`)
        self.txt_meri_num = QtWidgets.QLineEdit()
        self.txt_meri_num.setPlaceholderText("5")
        self.txt_meri_num.setValidator(QtGui.QIntValidator(1, 720))

        form = QtWidgets.QFormLayout(self.meridionals_box)
        form.addRow(self.opt_meri_auto)
        form.addRow(self.opt_meri_at, self.txt_meri_at)
        form.addRow(self.opt_meri_every, self.txt_meri_every)
        form.addRow(self.opt_meri_num, self.txt_meri_num)

        vbox.addWidget(self.meridionals_box)

        #: Group box for options specific to parallels (see
        #: :attr:`opt_para_auto`, :attr:`opt_para_at` and
        #: :attr:`opt_para_every`, :attr:`opt_para_num`)
        self.parallels_box = QtWidgets.QGroupBox('Parallels')
        self.parallels_box.setCheckable(True)

        #: Radiobutton for automatic drawing of parallels
        self.opt_para_auto = QtWidgets.QRadioButton("auto")

        #: Radiobutton for giving the exact position of parallels (see
        #: :attr:`txt_para_at`)
        self.opt_para_at = QtWidgets.QRadioButton("At:")

        #: Text field to enter the location of the parallels on the map (see
        #: :attr:`opt_para_at`)
        self.txt_para_at = QtWidgets.QLineEdit()
        self.txt_para_at.setPlaceholderText("-60, -30, 0, 30, ... °N")
        # TODO: Add validator

        #: Radiobutton for equal-width spaced parallels (see
        #: :attr:`txt_para_every`)
        self.opt_para_every = QtWidgets.QRadioButton("Every:")

        #: Text box to specify the distance between two parallels (see
        #: :attr:`opt_para_every`)
        self.txt_para_every = QtWidgets.QLineEdit()
        self.txt_para_every.setPlaceholderText("30 °N")
        self.txt_para_every.setValidator(QtGui.QDoubleValidator(0, 90, 7))

        #: Radiobutton to draw a specific number of parallels with
        #: equal-distance (see also :attr:`txt_para_num`)
        self.opt_para_num = QtWidgets.QRadioButton("Number:")

        #: Text box to set the number of parallels to be shown (see
        #: :attr:`opt_para_num`)
        self.txt_para_num = QtWidgets.QLineEdit()
        self.txt_para_num.setPlaceholderText("5")
        self.txt_para_num.setValidator(QtGui.QIntValidator(1, 360))

        form = QtWidgets.QFormLayout(self.parallels_box)
        form.addRow(self.opt_para_auto)
        form.addRow(self.opt_para_at, self.txt_para_at)
        form.addRow(self.opt_para_every, self.txt_para_every)
        form.addRow(self.opt_para_num, self.txt_para_num)

        vbox.addWidget(self.parallels_box)

        vbox.addWidget(self.button_box)

        self.fill_from_plotter(plotter)

        for button in [self.opt_meri_at, self.opt_meri_auto, self.opt_meri_num,
                       self.opt_meri_every, self.opt_para_at,
                       self.opt_para_auto, self.opt_para_num,
                       self.opt_para_every]:
            button.clicked.connect(self.update_forms)

    @property
    def default_colors(self) -> Dict[str, Color]:
        """Get default colors for the color labels in :attr:`widgets`."""
        import cartopy.feature as cf
        import matplotlib as mpl
        return {
            'background': mpl.rcParams['axes.facecolor'],
            'land': cf.LAND._kwargs['facecolor'],
            'ocean': cf.OCEAN._kwargs['facecolor'],
            'coast': 'k',
            }

    def get_colors(self, plotter: Plotter) -> Dict[str, Color]:
        """Get the colors for :attr:`widgets` from the plotter formatoptions.

        Parameters
        ----------
        plotter: psy_maps.plotters.MapPlotter
            The plotter with the formatoptions

        Returns
        -------
        dict
            A mapping from formatoptions in :attr:`colors` to the corresponding
            color in the `plotter`.
        """
        ret = {}
        if plotter.background.value != 'rc':
            ret['background'] = plotter.background.value
        lsm = plotter.lsm.value
        for part in ['land', 'ocean', 'coast']:
            if part in lsm:
                ret[part] = lsm[part]
        return ret

    def fill_from_plotter(self, plotter: Plotter) -> None:
        """Fill the dialog from a given plotter.

        Parameters
        ----------
        plotter: psy_maps.plotters.MapPlotter
            The plotter to get the formatoptions from.
        """
        chosen_colors = self.get_colors(plotter)

        for i, col in enumerate(self.colors):
            enable = col in chosen_colors
            cb = self.widgets.iloc[0, i]
            lbl = self.widgets.iloc[1, i]
            cb.setChecked(enable)
            if enable:
                lbl._set_color(chosen_colors[col])

        if plotter.clon.value is not None:
            self.txt_clon.setText(str(plotter.clon.value))
        if plotter.clat.value is not None:
            self.txt_clat.setText(str(plotter.clat.value))

        lsm = plotter.lsm.value

        if not lsm:
            self.lsm_box.setChecked(False)
        else:
            res = lsm['res']
            getattr(self, 'opt_' + res).setChecked(True)

        grid_labels = plotter.grid_labels.value
        if grid_labels is None:
            grid_labels = True
        self.grid_labels_box.setChecked(grid_labels)
        self.txt_grid_fontsize.setText(str(plotter.grid_labelsize.value))

        self.xgrid_value = None
        value = plotter.xgrid.value
        if not value:
            self.meridionals_box.setChecked(False)
        elif value is True:
            self.opt_meri_auto.setChecked(True)
        elif isinstance(value[0], str):
            self.xgrid_value = value[0]
            self.opt_meri_num.setChecked(True)
            self.txt_meri_num.setText(str(value[1]))
        elif isinstance(value, tuple):
            self.xgrid_value: Tuple[Any, Any] = value[:2]  # type: ignore
            self.opt_meri_num.setChecked(True)
            steps = 11 if len(value) == 2 else value[3]
            self.txt_meri_num.setText(str(steps))
        else:
            self.opt_meri_at.setChecked(True)
            self.txt_meri_at.setText(', '.join(map(str, value)))

        self.ygrid_value = None
        value = plotter.ygrid.value
        if not value:
            self.parallels_box.setChecked(False)
        elif value is True:
            self.opt_para_auto.setChecked(True)
        elif isinstance(value[0], str):
            self.opt_para_num.setChecked(True)
            self.txt_para_num.setText(str(value[1]))
            self.ygrid_value = value[0]
        elif isinstance(value, tuple):
            self.ygrid_value: Tuple[Any, Any] = value[:2]  # type: ignore
            self.opt_para_num.setChecked(True)
            steps = 11 if len(value) == 2 else value[3]
            self.txt_para_num.setText(str(steps))
        else:
            self.opt_para_at.setChecked(True)
            self.txt_para_at.setText(', '.join(map(str, value)))

    def update_forms(self) -> None:
        """Update text widgets for the options to draw merdionals and parallels.
        """
        if self.meridionals_box.isChecked():
            self.txt_meri_at.setEnabled(self.opt_meri_at.isChecked())
            self.txt_meri_every.setEnabled(self.opt_meri_every.isChecked())
            self.txt_meri_num.setEnabled(self.opt_meri_num.isChecked())
        if self.parallels_box.isChecked():
            self.txt_para_at.setEnabled(self.opt_para_at.isChecked())
            self.txt_para_every.setEnabled(self.opt_para_every.isChecked())
            self.txt_para_num.setEnabled(self.opt_para_num.isChecked())

    @property
    def value(self) -> Dict[str, Any]:
        """Get the formatoptions of this dialog to update a plotter."""
        import numpy as np
        ret: Dict[str, Any] = {}
        ret['clon'] = None if not self.txt_clon.text().strip() else float(
            self.txt_clon.text().strip())
        ret['clat'] = None if not self.txt_clat.text().strip() else float(
            self.txt_clat.text().strip())

        lsm: LSM_T = {}

        for col in ['land', 'ocean', 'coast']:
            lbl = self.widgets.loc['color', col]
            if lbl.isEnabled():
                lsm[col] = list(lbl.color.getRgbF())

        if lsm or self.lsm_box.isChecked():
            if self.opt_110m.isChecked():
                lsm['res'] = '110m'
            elif self.opt_50m.isChecked():
                lsm['res'] = '50m'
            elif self.opt_10m.isChecked():
                lsm['res'] = '10m'
            else:
                lsm['res'] = '110m'
        else:
            lsm['res'] = False
        if lsm:
            ret['lsm'] = lsm

        bc_lbl = self.widgets.loc['color', 'background']
        if bc_lbl.isEnabled():
            ret['background'] = list(bc_lbl.color.getRgbF())

        ret["grid_labels"] = self.grid_labels_box.isChecked()
        if ret["grid_labels"]:
            ret["grid_labels"] = None
            labelsize = self.txt_grid_fontsize.text().strip()
            if labelsize:
                try:
                    labelsize = float(labelsize)
                except TypeError:
                    pass
                ret["grid_labelsize"] = labelsize

        if not self.meridionals_box.isChecked():
            ret['xgrid'] = False
        elif self.opt_meri_auto.isChecked():
            ret['xgrid'] = True
        elif self.opt_meri_every.isChecked():
            ret['xgrid'] = np.arange(
                -180, 180, float(self.txt_meri_every.text().strip() or 30))
        elif self.opt_meri_at.isChecked():
            ret['xgrid'] = list(map(
                float, self.txt_meri_at.text().split(','))) or False
        elif self.opt_meri_num.isChecked():
            if self.xgrid_value is None:
                ret['xgrid'] = ['rounded', int(self.txt_meri_num.text() or 5)]
            elif isinstance(self.xgrid_value, str):
                ret['xgrid'] = [self.xgrid_value,
                                int(self.txt_meri_num.text() or 5)]
            else:
                ret['xgrid'] = tuple(self.xgrid_value) + (
                    int(self.txt_meri_num.text() or 5), )

        if not self.parallels_box.isChecked():
            ret['ygrid'] = False
        elif self.opt_para_auto.isChecked():
            ret['ygrid'] = True
        elif self.opt_para_every.isChecked():
            ret['ygrid'] = np.arange(
                -180, 180, float(self.txt_para_every.text().strip() or 30))
        elif self.opt_para_at.isChecked():
            ret['ygrid'] = list(map(
                float, self.txt_para_at.text().split(','))) or False
        elif self.opt_para_num.isChecked():
            if self.ygrid_value is None:
                ret['ygrid'] = ['rounded', int(self.txt_para_num.text() or 5)]
            elif isinstance(self.ygrid_value, str):
                ret['ygrid'] = [self.ygrid_value,
                                int(self.txt_para_num.text() or 5)]
            else:
                ret['ygrid'] = tuple(self.ygrid_value) + (
                    int(self.txt_para_num.text() or 5), )
        return ret

    @classmethod
    def update_plotter(cls, plotter: Plotter) -> None:
        """Open a :class:`BasemapDialog` to update a plotter.

        Parameters
        ----------
        plotter: psy_maps.plotters.MapPlotter
            The plotter to update.
        """
        dialog = cls(plotter)
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()
        dialog.exec_()
        if dialog.result() == QtWidgets.QDialog.Accepted:
            plotter.update(
                **dialog.value)


class CmapDialog(QtWidgets.QDialog):
    """A dialog to modify color bounds and colormaps."""

    def __init__(self, project: Project, *args, **kwargs) -> None:
        """
        Parameters
        ----------
        project: psyplot.project.Project
            The psyplot project to update. Note that we will only use the
            very first plotter in this project
        """
        import psy_simple.widgets.colors as pswc
        super().__init__(*args, **kwargs)

        #: Button box to accept or cancel this dialog
        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            self)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)

        #: Mapping from formatoption key to :class:`LabelWidgetLine` widgets to
        #: controlling the formatoption
        self.fmt_widgets = {}
        plotter = project(fmts=['cmap', 'bounds']).plotters[0]

        #: Widget for manipulating the color map
        self.cmap_widget = self.fmt_widgets['cmap'] = LabelWidgetLine(
            plotter.cmap, project, pswc.CMapFmtWidget,
            widget_kws=dict(properties=False))
        self.cmap_widget.editor.setVisible(False)
        self.cmap_widget.editor.line_edit.textChanged.connect(
            self.update_preview)

        #: tabs for switching between bounds (:attr:`bounds_widget`) and
        #: colorbar ticks (:attr:`cticks_widget`)
        self.tabs = QtWidgets.QTabWidget()

        #: :class:`LabelWidgetLine` to controll the colorbar bounds
        self.bounds_widget = self.fmt_widgets['bounds'] = LabelWidgetLine(
            plotter.bounds, project, pswc.BoundsFmtWidget,
            widget_kws=dict(properties=False))
        self.bounds_widget.editor.line_edit.textChanged.connect(
            self.update_preview)
        self.tabs.addTab(self.bounds_widget, "Colormap boundaries")

        #: :class:`LabelWidgetLine` to controll the ctick positions
        self.cticks_widget = self.fmt_widgets['cticks'] = LabelWidgetLine(
            plotter.cticks, project, pswc.CTicksFmtWidget,
            widget_kws=dict(properties=False))
        self.cticks_widget.editor.line_edit.textChanged.connect(
            self.update_preview)
        self.tabs.addTab(self.cticks_widget, "Colorbar ticks")

        #: :class:`ColorbarPreview` to show a preview of the colorbar with
        #: the selected formatoption in :attr:`fmt_widgets`
        self.cbar_preview = ColorbarPreview(plotter)
        self.cbar_preview.setMaximumHeight(self.tabs.sizeHint().height() // 3)

        vbox = QtWidgets.QVBoxLayout(self)
        vbox.addWidget(self.cmap_widget)
        vbox.addWidget(self.tabs)
        vbox.addWidget(self.cbar_preview)
        vbox.addWidget(self.button_box)

    @property
    def plotter(self) -> Plotter:
        """Get the plotter with the formatoptions we use to fill this dialog."""
        return self.bounds_widget.editor.fmto.plotter

    def update_preview(self) -> None:
        """Update the :attr:`cbar_preview` from the various :attr:`fmt_widgets`.
        """
        try:
            bounds = self.bounds_widget.editor.value
        except Exception:
            bounds = self.plotter.bounds.value
        try:
            cticks = self.cticks_widget.editor.value
        except Exception:
            cticks = self.plotter.cticks.value
        try:
            cmap = self.cmap_widget.editor.value
        except Exception:
            cmap = self.plotter.cmap.value
        self.cbar_preview.update_colorbar(
            bounds=bounds, cticks=cticks, cmap=cmap)

    @property
    def fmts(self) -> Dict[str, Any]:
        """Map from formatoption in :attr:`fmt_widgets` to values."""
        ret = {}
        for fmt, widget in self.fmt_widgets.items():
            if widget.editor.changed:
                try:
                    value = widget.editor.value
                except:
                    raise IOError(f"{fmt}-value {widget.editor.text} could "
                                  "not be parsed to python!")
                else:
                    ret[fmt] = value
        return ret

    @classmethod
    def update_project(cls, project: Project) -> None:
        """Create a :class:`CmapDialog` to update a `project`

        This classmethod creates a new :class:`CmapDialog` instance, fills it
        with the formatoptions of the first plotter in `project`, enters the
        main event loop, and updates the `project` upon acceptance.

        Parameters
        ----------
        project: psyplot.project.Project
            The psyplot project to update
        """
        dialog = cls(project)
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()
        dialog.exec_()
        if dialog.result() == QtWidgets.QDialog.Accepted:
            project.update(**dialog.fmts)


class _DummyFormatOption(Formatoption):
    """Dummy formatoption for static type checking.
    
    This is just a workaround for the static type checker to be able to tell
    what the :class:`FakePlotter` formatoptions are, used in 
    :attr:`ColorbarPreview.fake_plotter`
    """
    def update(self,):
        pass


class FakePlotter(Plotter):
    """A dummy plotter for the colorbar preview."""

    bounds: Formatoption = _DummyFormatOption('bounds')
    cmap: Formatoption = _DummyFormatOption('cmap')
    cticks: Formatoption = _DummyFormatOption('cticks')
    cbar: Formatoption = _DummyFormatOption('cbar')


class ColorbarPreview(FigureCanvas):
    """A preview widget of a colorbar.

    This matplotlib figure contains one single axes to display the colorbar
    filled by the formatoptions of a given `plotter`."""

    def __init__(
            self,
            plotter: Plotter,
            parent: Optional[QtWidgets.QWidget] = None,
            *args, **kwargs
        ) -> None:
        """
        Parameters
        ----------
        plotter: psy_simple.plotters.Base2D
            The plotter to use to draw the colorbar from
        parent: QtWidget.QWidget
            The parent widget
        """
        fig = Figure(*args, **kwargs)

        FigureCanvas.__init__(self, fig)
        self.setParent(parent)

        FigureCanvas.setSizePolicy(self,
                                   QtWidgets.QSizePolicy.Expanding,
                                   QtWidgets.QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        self.axes_counter = 0

        #: The plotter to use for displaying the colorbar
        self.plotter = plotter
        self.init_colorbar(plotter)

    def resizeEvent(self, event: QEvent) -> Any:
        """Reimplemented to make sure we cannot get smaller than 0."""
        h = event.size().height()
        if h <= 0:
            return
        return super().resizeEvent(event)

    def init_colorbar(self, plotter: Plotter) -> None:
        """Initialize the colorbar.

        This method extracts the formatoptions of the given `plotter` and draws
        the colorbar.
        """
        from matplotlib.cm import ScalarMappable
        norm = plotter.bounds.norm
        cmap = plotter.cmap.get_cmap(self.plotter.plot.array)

        self.mappable = sm = ScalarMappable(cmap=cmap, norm=norm)
        sm.set_array([])

        self.cax = self.figure.add_axes([0.1, 0.5, 0.8, 0.5],
                                        label=self.axes_counter)

        self.cbar = self.figure.colorbar(
            sm, norm=norm, cmap=cmap, cax=self.cax, orientation='horizontal')

    @property
    def fake_plotter(self) -> FakePlotter:
        """Create a plotter with the formatoptions of the real :attr:`plotter`.

        We can update this plotter without impacting the origin :attr:`plotter`
        """
        from psyplot.plotter import Plotter

        class _FakePlotter(FakePlotter):
            bounds = self.plotter.bounds.__class__('bounds')
            cmap = self.plotter.cmap.__class__('cmap')
            cticks = self.plotter.cticks.__class__('cticks')
            cbar = self.plotter.cbar.__class__('cbar')

            _rcparams_string = self.plotter._get_rc_strings()

        ref = self.plotter
        fig = Figure()
        ax = fig.add_subplot()

        plotter = _FakePlotter(
            ref.data.copy(), make_plot=False, bounds=ref['bounds'],
            cmap=ref['cmap'], cticks=ref['cticks'], cbar='', ax=ax)

        plotter.cticks._colorbar = self.cbar

        plotter.plot_data = ref.plot_data
        return plotter

    def update_colorbar(self, **kwargs) -> None:
        """Update the colorbar with new formatoptions.

        This method takes the :attr:`fake_plotter`, updates it from the given
        `kwargs`, updates the colorbar preview.

        Parameters
        ----------
        ``**kwargs``
            `bounds`, `cmap`, `cticks` or `cbar` formatoption keyword-value
            pairs
        """

        # create a dummy plotter
        plotter = self.fake_plotter

        # update from the given kwargs
        try:
            for key, val in kwargs.items():
                plotter[key] = val
        except (ValueError, TypeError):
            return

        plotter.initialize_plot(ax=plotter.ax)

        current_norm = self.mappable.norm
        current_cmap = self.mappable.get_cmap()
        current_locator = self.cbar.locator

        # update the preview with the norm of the plotter
        try:
            try:
                plotter.bounds.norm._check_vmin_vmax()
            except (AttributeError, TypeError):
                pass
            try:
                plotter.bounds.norm.autoscale_None(plotter.bounds.array)
            except AttributeError:
                pass
            self.mappable.set_norm(plotter.bounds.norm)
            self.mappable.set_cmap(plotter.cmap.get_cmap(
                self.plotter.plot.array))
            plotter.cticks.colorbar = self.cbar
            plotter.cticks.default_locator = \
                self.plotter.cticks.default_locator
            plotter.cticks.update_axis(plotter.cticks.value)
            self.draw()

        except Exception:
            self.mappable.set_norm(current_norm)
            self.mappable.set_cmap(current_cmap)
            self.cbar.locator = current_locator
            self.cbar.update_ticks()


class FormatoptionsEditor(QtWidgets.QWidget):
    """A widget to update a formatoption.

    This widget is a light-weight version of the
    :class:`psyplot_gui.fmt_widget.FormatoptionsWidget` class. It contains
    a line editor and a text editor to set the value of a specific formatoption.
    """

    def __init__(self, fmto: Formatoption, *args, **kwargs) -> None:
        """
        Parameters
        ----------
        fmto: psyplot.plotter.Formatoption
            The formatoption instance to display the value from
        """
        super().__init__(*args, **kwargs)
        layout = QtWidgets.QHBoxLayout()

        #: The :class:`~psyplot.plotter.Formatoption` that fills this widget
        self.fmto = fmto

        #: A single line editor holding the formatoption value (see also
        #: :attr:`text_edit` and :attr:`btn_multiline`)
        self.line_edit = QtWidgets.QLineEdit()
        layout.addWidget(self.line_edit)

        #: A multi-line editor holiding the value of :attr:`fmto` (see also
        #: :attr:`line_edit` and :attr:`btn_multiline`)
        self.text_edit = QtWidgets.QTextEdit()
        self.text_edit.setVisible(False)
        layout.addWidget(self.text_edit)

        #: A tool button to switch from the single line editor :attr:`line_edit`
        #: to the multi-line editor :attr:`text_edit`
        self.btn_multiline = QtWidgets.QToolButton()
        self.btn_multiline.setText('⌵')
        self.btn_multiline.setCheckable(True)
        self.btn_multiline.setToolTip("Toggle multiline editor")
        self.btn_multiline.clicked.connect(self.toggle_multiline)
        layout.addWidget(self.btn_multiline)

        self.insert_obj(fmto.value)

        #: Value of the :attr:`fmto` at the initialization of this widget
        self.initial_value = self.line_edit.text()

        self.setLayout(layout)

    @property
    def changed(self) -> bool:
        """Check if the value in this editor differs from the original `fmto`.
        """
        return self.fmto.diff(self.fmto.validate(self.get_obj()))

    def toggle_multiline(self) -> None:
        """Switch from :attr:`line_edit` and :attr:`text_edit` or back."""
        multiline = self.multiline
        self.text_edit.setVisible(multiline)
        self.line_edit.setVisible(not multiline)
        if multiline:
            self.text_edit.setPlainText(self.line_edit.text())
        else:
            self.line_edit.setText(self.text_edit.toPlainText())

    @property
    def multiline(self) -> bool:
        """True if the :attr:`text_edit` should be visible."""
        return self.btn_multiline.isChecked()

    @property
    def text(self) -> str:
        """Text of the :attr:`text_edit` (or :attr:`line_edit`)."""
        return (self.text_edit.toPlainText() if self.multiline else
                self.line_edit.text())

    @text.setter
    def text(self, s: str) -> None:
        self.clear_text()
        if self.multiline:
            self.text_edit.insertPlainText(s)
        else:
            self.line_edit.insert(s)

    @property
    def value(self) -> Any:
        """Load the value of :attr:`text` with yaml."""
        text = self.text
        return yaml.load(text, Loader=yaml.Loader)

    def clear_text(self) -> None:
        """Clear the editor."""
        if self.multiline:
            self.text_edit.clear()
        else:
            self.line_edit.clear()

    def set_obj(self, obj: Any) -> None:
        """Clear the editor and set another object."""
        self.clear_text()
        self.insert_obj(obj)

    def get_obj(self) -> Any:
        """Alias for :attr:`value`."""
        return self.value

    def insert_obj(self, obj: Any) -> None:
        """Add a string to the formatoption widget.

        Parameters
        ----------
        obj: object
            The object to insert into the line editor. it will be dumped
            using yaml and displayed in the :attr:`text_edit` (or
            :attr:`line_edit`)
        """
        current = self.text
        use_line_edit = not self.multiline
        # strings are treated separately such that we consider quotation marks
        # at the borders
        if isinstance(obj, str) and current:
            if use_line_edit:
                pos = self.line_edit.cursorPosition()
            else:
                pos = self.text_edit.textCursor().position()
            if pos not in [0, len(current)]:
                s = obj
            else:
                if current[0] in ['"', "'"]:
                    current = current[1:-1]
                self.clear_text()
                if pos == 0:
                    s = '"' + obj + current + '"'
                else:
                    s = '"' + current + obj + '"'
                current = ''
        elif isinstance(obj, str):  # add quotation marks
            s = '"' + obj + '"'
        else:
            s = yaml.dump(obj, default_flow_style=True).strip()
            if s.endswith('\n...'):
                s = s[:-4]
        if use_line_edit:
            self.line_edit.insert(s)
        else:
            self.text_edit.insertPlainText(s)


class LabelWidgetLine(QtWidgets.QGroupBox):
    """A widget to change a formatoption.

    This class holds a :class:`FormatoptionsEditor` to control the
    appearance of a specific formatoption. Additionally it displays the
    formatoption specific line widget (see
    :meth:`psyplot.plotter.Formatoption.get_fmt_widget`) to contol it.
    """

    def __init__(self, fmto: Formatoption, project: Project,
                 fmto_widget: Type[QtWidgets.QWidget],
                 widget_kws: Dict[str, Any] = {}, *args, **kwargs) -> None:
        """
        Parameters
        ----------
        fmto: psyplot.plotter.Formatoption
            The formatoption to manipulate
        project: psyplot.project.Project
            The project to use to fill this formatoption
        fmto_widget: type
            A subclass of the :class:`QWidget` that can be used to control
            the formatoption. This class is commonly used in the
            :meth:`psyplot.plotter.Formatoption.get_fmt_widget` of the given
            `fmto`
        widget_kws: dict
            Further keywords that are passed to the creation of the
            `fmto_widget` instance.
        """
        super().__init__(f'{fmto.name} ({fmto.key})', *args, **kwargs)
        self.editor = FormatoptionsEditor(fmto)
        vbox = QtWidgets.QVBoxLayout()
        vbox.addWidget(
            fmto_widget(self.editor, fmto, project, **widget_kws))
        vbox.addWidget(self.editor)
        self.setLayout(vbox)


class LabelDialog(QtWidgets.QDialog):
    """A dialog to change labels.

    This class contains one :class:`LabelWidgetLine` per text formatoption."""

    def __init__(self, project: Project, *fmts: str) -> None:
        """
        Parameters
        ----------
        project: psyplot.project.Project
            The psyplot project to update. Note that we will only use the
            very first plotter in this project
        ``*fmts``
            The formatoption keys to display. Each formatoption should be a
            subclass of :class:`psy_simple.base.TextBase`
        """
        from psy_simple.widgets.texts import LabelWidget
        super().__init__()
        self.project = project
        layout = QtWidgets.QVBoxLayout()
        plotter = project.plotters[0]
        self.fmt_widgets = {}
        for fmt in fmts:
            fmto = getattr(plotter, fmt)
            fmt_widget = LabelWidgetLine(
                fmto, project, LabelWidget, widget_kws=dict(properties=False))
            self.fmt_widgets[fmt] = fmt_widget
            layout.addWidget(fmt_widget)

        self.button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel,
            self)
        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.reject)
        layout.addWidget(self.button_box)
        self.setLayout(layout)

    @property
    def fmts(self) -> Dict[str, Any]:
        """Mapping from formatoption key to value in this dialog."""
        ret = {}
        for fmt, widget in self.fmt_widgets.items():
            if widget.editor.changed:
                try:
                    value = widget.editor.value
                except Exception:
                    raise IOError(f"{fmt}-value {widget.editor.text} could "
                                  "not be parsed to python!")
                else:
                    ret[fmt] = value
        return ret

    @classmethod
    def update_project(cls, project: Project, *fmts: str) -> None:
        """Create a :class:`LabelDialog` to update the labels in a `project`.

        This classmethod creates a new :class:`LabelDialog` instance, fills it
        with the formatoptions of the first plotter in `project`, enters the
        main event loop, and updates the `project` upon acceptance.

        Parameters
        ----------
        project: psyplot.project.Project
            The psyplot project to update. Note that we will only use the
            very first plotter in this project
        ``*fmts``
            The formatoption keys to display. Each formatoption should be a
            subclass of :class:`psy_simple.base.TextBase`
        """
        dialog = cls(project, *fmts)
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()
        dialog.exec_()
        if dialog.result() == QtWidgets.QDialog.Accepted:
            project.update(
                **dialog.fmts)



# -*- coding: utf-8 -*-
"""ncview-like GUI to the psyplot framework

**Disclaimer**

Copyright (C) 2020 Philipp S. Sommer

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see https://www.gnu.org/licenses/."""
from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    Dict,
    Any,
    Optional,
    Tuple,
    Union,
    Type,
)

import sys
import argparse

if TYPE_CHECKING:
    from xarray import Dataset

from ._version import get_versions

__version__ = get_versions()['version']

del get_versions

__author__ = "Philipp S. Sommer"

__copyright__ = "Copyright 2020, Philipp S. Sommer"

__email__ = "philipp.sommer@hzg.de"

__status__ = "Development"

__license__ = "GPLv3"


def start_app(
        ds: Optional[Dataset], name: Optional[str] = None,
        plotmethod: str = 'mapplot', preset: Optional[str] = None
    ) -> None:
    """Start the standalone GUI application.

    This function creates a `QApplication` instance, an instance of the
    :class:`psy_view.ds_widget.DatasetWidget` and enters the main event loop.

    Parameters
    ----------
    ds: xarray.Dataset
        The dataset to display. If None, the user can select it afterwards
    name: str
        The variable name in `ds` to display. If None, the user can select it
        afterwards
    plotmethod: {'mapplot' | 'lineplot' | 'plot2d' }
        The plotmethod to use
    preset: str
        The preset to apply
    """
    from PyQt5 import QtWidgets
    from PyQt5.QtGui import QIcon  # pylint: disable=no-name-in-module
    from psyplot_gui import rcParams

    rcParams['help_explorer.use_webengineview'] = False

    from psy_view.ds_widget import DatasetWidget
    from psyplot_gui.common import get_icon

    app = QtWidgets.QApplication(sys.argv)
    ds_widget = DatasetWidget(ds)
    ds_widget.setWindowIcon(QIcon(get_icon('logo.svg')))
    if preset is not None:
        ds_widget.load_preset(preset)
    if name is not None:
        if ds is None:
            raise ValueError("Variable specified but without dataset")
        elif name not in ds_widget.variable_buttons:
            valid = list(ds_widget.variable_buttons)
            raise ValueError(f"{name} is not part of the dataset. "
                             f"Possible variables are {valid}.")
        ds_widget.plotmethod = plotmethod
        ds_widget.variable = name
        ds_widget.make_plot()
        ds_widget.refresh()
    ds_widget.show()
    ds_widget.show_current_figure()
    sys.excepthook = ds_widget.excepthook
    sys.exit(app.exec_())


def get_parser() -> argparse.ArgumentParser:
    """Get the command line parser for psy-view."""
    from textwrap import dedent
    parser = argparse.ArgumentParser('psy-view')

    parser.add_argument(
        'input_file', help="The file to visualize", nargs='?', default=None)

    parser.add_argument(
        '-n', '--name',
        help=("Variable name to display. Don't provide a variable to display "
              "the first variable found in the dataset."),
        const=object, nargs="?")

    parser.add_argument(
        '-pm', '--plotmethod', help="The plotmethod to use", default="mapplot",
        choices=["mapplot", "plot2d", "lineplot"])

    parser.add_argument(
        '--preset', help="Apply a preset to the plot")

    parser.add_argument(
        '-V', '--version', action='version', version=__version__)

    parser.epilog = dedent("""
    psy-view  Copyright (C) 2020  Philipp S. Sommer

    This program comes with ABSOLUTELY NO WARRANTY.
    This is free software, and you are welcome to redistribute it
    under the conditions of the GNU GENERAL PUBLIC LICENSE, Version 3.""")

    return parser


def main() -> None:
    """Start the app with the provided command-line options."""
    import psyplot.project as psy
    parser = get_parser()
    args = parser.parse_known_args()[0]

    if args.input_file is not None:
        try:
            ds = psy.open_dataset(args.input_file)
        except:
            ds = psy.open_dataset(args.input_file, decode_times=False)
    else:
        ds = None

    if args.name is object and ds is not None:
        args.name = list(ds)[0]

    start_app(ds, args.name, args.plotmethod, args.preset)


if __name__ == "__main__":
    main()



"""Plotmethod widgets.

This module defines the widgets to interface with the mapplot, plot2d and
lineplot plotmethods

**Disclaimer**

Copyright (C) 2020 Philipp S. Sommer

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.If not, see https: // www.gnu.org / licenses / . """
from __future__ import annotations
import os.path as osp

from typing import (
    TYPE_CHECKING,
    ClassVar,
    Callable,
    Optional,
    Union,
    List,
    Hashable,
    Dict,
    Any,
    Tuple,
    Iterator,
)

from functools import partial
from itertools import chain, cycle
import contextlib
import textwrap

import xarray as xr
from psyplot.utils import unique_everseen

from PyQt5 import QtWidgets, QtCore, QtGui
import psy_view.dialogs as dialogs
import psy_view.utils as utils
from psy_view.rcsetup import rcParams

from psyplot_gui.common import get_icon as get_psy_icon
import psy_simple.widgets.colors as pswc
import matplotlib.colors as mcol

if TYPE_CHECKING:
    from xarray import DataArray, Dataset, Variable
    from psyplot.project import PlotterInterface, Project
    from psyplot.data import InteractiveList
    from psyplot.plotter import Plotter


class PlotMethodWidget(QtWidgets.QWidget):
    """Base class for interfacing a psyplot plotmethod.

    This method serves as a base class for interfacing any of the psyplot
    plot methods registered via :func:`psyplot.project.register_plotter`.

    The name of the plotmethod should be implemented as the :attr:`plotmethod`
    attribute.
    """

    plotmethod: ClassVar[str] = ''

    #: trigger a replot of this widget. This can be emitted with the
    #: :meth:`trigger_replot` method
    replot = QtCore.pyqtSignal(str)

    #: trigger a replot of this widget. This can be emitted with the
    #: :meth:`trigger_reset` method
    reset = QtCore.pyqtSignal(str)

    #: signalize that the widget has been changed but not plot changes are
    #: needed
    changed = QtCore.pyqtSignal(str)

    array_info = None


    layout: QtWidgets.QLayout = None

    def __init__(
            self, get_sp: Callable[[], Optional[Project]],
            ds: Optional[Dataset]):
        super().__init__()
        self._get_sp = get_sp

        self.setup()

        if hasattr(self, "layout"):
            self.setLayout(self.layout)

        self.refresh(ds)

    def setup(self):
        """Set up widget during initialization."""
        pass

    @property
    def sp(self) -> Optional[Project]:
        """Get the subproject of this plotmethod interface."""
        return getattr(self._get_sp(), self.plotmethod, None)

    @property
    def data(self) -> Union[DataArray, InteractiveList]:
        """Get the data of this plotmethod interface."""
        if self.sp is None:
            raise ValueError("No plot has yet been initialized")
        else:
            return self.sp[0]

    @property
    def plotter(self) -> Optional[Plotter]:
        """Get the first plotter of the :attr:`sp` project."""
        if self.sp and self.sp.plotters:
            return self.sp.plotters[0]
        else:
            return None

    @property
    def formatoptions(self) -> List[str]:
        """Get the formatoption keys of this plotmethod."""
        if self.plotter is not None:
            return list(self.plotter)
        else:
            import psyplot.project as psy
            return list(getattr(psy.plot, self.plotmethod).plotter_cls())

    def get_fmts(
            self, var: DataArray, init: bool = False
        ) -> Dict[Union[Hashable, str, Any], Any]:
        """Get the formatoptions for a new plot.
        
        Parameters
        ----------
        var: xarray.Variable
            The variable in the base dataset
        init: bool
            If True, call the :meth:`init_dims` method to inject necessary
            formatoptions and dimensions for the initialization.

        Returns
        -------
        dict
            A mapping from formatoption or dimension to the corresponding value
            for the plotmethod.
        """
        ret = {}
        if init:
            ret.update(self.init_dims(var))
        return ret

    def init_dims(
            self, var: DataArray
        ) -> Dict[Union[Hashable, str, Any], Any]:
        """Get the formatoptions for a new plot.
        
        Parameters
        ----------
        var: xarray.Variable
            The variable in the base dataset

        Returns
        -------
        dict
            A mapping from formatoption or dimension to the corresponding value
            for the plotmethod.
        
        """
        return {}

    def refresh(self, ds: Optional[Dataset]) -> None:
        """Refresh this widget from the given dataset."""
        self.setEnabled(bool(self.sp))

    def trigger_replot(self) -> None:
        """Emit the :attr:`replot` signal to replot the project."""
        self.replot.emit(self.plotmethod)

    def trigger_reset(self):
        """Emit the :attr:`reset` signal to reinitialize the project."""
        self.array_info = self.sp.array_info(
            standardize_dims=False)[self.sp[0].psy.arr_name]
        self.reset.emit(self.plotmethod)

    def trigger_refresh(self):
        """Emit the :attr:`changed` signal to notify changes in the plot."""
        self.changed.emit(self.plotmethod)

    def get_slice(
            self, x: float, y: float
        ) -> Optional[Dict[Hashable, Union[int, slice]]]:
        """Get the slice for the selected coordinates.

        This method is called when the user clicks on the coordinates in the 
        plot.

        See Also
        --------
        psy_view.ds_widget.DatasetWidget.display_line

        Notes
        -----
        This is reimplemented in the :class:`MapPlotWidget`.
        """
        return None

    def valid_variables(self, ds: Dataset) -> List[Hashable]:
        """Get a list of variables that can be visualized with this plotmethod.

        Parameters
        ----------
        ds: xarray.Dataset
            The dataset to use

        Returns
        -------
        list of str
            List of variable names to plot
        """
        ret = []
        plotmethod = getattr(ds.psy.plot, self.plotmethod)
        for v in list(ds):
            init_kws = self.init_dims(ds[v])  # type: ignore
            dims = init_kws.get('dims', {})
            decoder = init_kws.get('decoder')
            if plotmethod.check_data(ds, v, dims, decoder)[0][0]:
                ret.append(v)
        return ret


class MapPlotWidget(PlotMethodWidget):
    """A widget to control the mapplot plotmethod."""

    plotmethod = 'mapplot'

    def setup(self) -> None:
        """Set up color and projection buttons.
        
        See Also
        --------
        setup_color_buttons
        setup_projection_buttons"""
        self.layout = vbox = QtWidgets.QVBoxLayout()

        self.formatoptions_box = QtWidgets.QHBoxLayout()
        self.setup_color_buttons()
        self.setup_projection_buttons()
        self.btn_labels = utils.add_pushbutton(
            "Labels", self.edit_labels, "Edit title, colorbar labels, etc.",
            self.formatoptions_box)

        vbox.addLayout(self.formatoptions_box)

        self.dimension_box = QtWidgets.QGridLayout()
        self.setup_dimension_box()

        vbox.addLayout(self.dimension_box)

    def setup_color_buttons(self) -> None:
        """Set up the buttons to change the colormap, etc."""
        self.btn_cmap = pswc.CmapButton()
        self.btn_cmap.setToolTip("Select a different colormap")
        self.formatoptions_box.addWidget(self.btn_cmap)
        self.btn_cmap.colormap_changed.connect(self.set_cmap)
        self.btn_cmap.colormap_changed[mcol.Colormap].connect(self.set_cmap)
        self.setup_cmap_menu()

        self.btn_cmap_settings = utils.add_pushbutton(
            utils.get_icon('color_settings'), self.edit_color_settings,
            "Edit color settings", self.formatoptions_box,
            icon=True)

    def setup_cmap_menu(self) -> QtWidgets.QMenu:
        """Set up the menu to change the colormaps."""
        menu = self.btn_cmap.cmap_menu

        menu.addSeparator()
        self.select_cmap_action = menu.addAction(
            'More colormaps', self.open_cmap_dialog)

        self.color_settings_action = menu.addAction(
            QtGui.QIcon(utils.get_icon('color_settings')), 'More options',
            self.edit_color_settings)

        return menu

    def open_cmap_dialog(self, N: int = 10) -> None:
        """Open the dialog to change the colormap.
        
        Parameters
        ----------
        N: int
            The number of colormaps to show

        See Also
        --------
        psy_simple.widgets.colors.CmapButton
        """
        if self.plotter:
            N = self.plotter.plot.mappable.get_cmap().N
        else:
            N = 10
        self.btn_cmap.open_cmap_dialog(N)

    def setup_projection_menu(self) -> QtWidgets.QMenu:
        """Set up the menu to modify the basemap."""
        menu = QtWidgets.QMenu()
        for projection in rcParams['projections']:
            menu.addAction(
                projection, partial(self.set_projection, projection))
        menu.addSeparator()
        self.proj_settings_action = menu.addAction(
            QtGui.QIcon(utils.get_icon('proj_settings')),
            "Customize basemap...", self.edit_basemap_settings)
        return menu

    def setup_projection_buttons(self) -> None:
        """Set up the buttons to modify the basemap."""
        self.btn_proj = utils.add_pushbutton(
            rcParams["projections"][0], self.choose_next_projection,
            "Change the basemap projection", self.formatoptions_box,
            toolbutton=True)
        self.btn_proj.setMenu(self.setup_projection_menu())
        max_width = max(map(self.btn_proj.fontMetrics().width,
                            rcParams['projections'])) * 2
        self.btn_proj.setMinimumWidth(max_width)
        self.btn_proj.setPopupMode(QtWidgets.QToolButton.MenuButtonPopup)

        self.btn_proj_settings = utils.add_pushbutton(
            utils.get_icon('proj_settings'), self.edit_basemap_settings,
            "Edit basemap settings", self.formatoptions_box,
            icon=True)

        self.btn_datagrid = utils.add_pushbutton(
            "Cells", self.toggle_datagrid,
            "Show the grid cell boundaries", self.formatoptions_box)
        self.btn_datagrid.setCheckable(True)

    def setup_dimension_box(self) -> None:
        """Set up a box to control, what is the x and y-dimension."""
        self.dimension_box = QtWidgets.QGridLayout()

        self.dimension_box.addWidget(QtWidgets.QLabel('x-Dimension:'), 0, 0)
        self.combo_xdim = QtWidgets.QComboBox()
        self.dimension_box.addWidget(self.combo_xdim, 0, 1)

        self.dimension_box.addWidget(QtWidgets.QLabel('y-Dimension:'), 0, 2)
        self.combo_ydim = QtWidgets.QComboBox()
        self.dimension_box.addWidget(self.combo_ydim, 0, 3)

        self.dimension_box.addWidget(QtWidgets.QLabel('x-Coordinate:'), 1, 0)
        self.combo_xcoord = QtWidgets.QComboBox()
        self.dimension_box.addWidget(self.combo_xcoord, 1, 1)

        self.dimension_box.addWidget(QtWidgets.QLabel('y-Coordinate:'), 1, 2)
        self.combo_ycoord = QtWidgets.QComboBox()
        self.dimension_box.addWidget(self.combo_ycoord, 1, 3)

        self.combo_xdim.currentTextChanged.connect(self.set_xcoord)
        self.combo_ydim.currentTextChanged.connect(self.set_ycoord)

        for combo in self.coord_combos:
            combo.currentIndexChanged.connect(self.trigger_refresh)

    def set_xcoord(self, text: str) -> None:
        """Set the name of the x-coordinate."""
        self.set_combo_text(self.combo_xcoord, text)

    def set_ycoord(self, text: str) -> None:
        """Set the name of the y-coordinate."""
        self.set_combo_text(self.combo_ycoord, text)

    def set_combo_text(self, combo: QtWidgets.QComboBox, text: str) -> None:
        """Convenience function to update set the current text of a combobox.
        
        Parameters
        ----------
        combo: PyQt5.QtWidgets.QComboBox
            The combobox to modify
        text: str
            The item to use"""
        items = list(map(combo.itemText, range(combo.count())))
        if text in items:
            combo.setCurrentIndex(items.index(text))

    def init_dims(
            self, var: DataArray
        ) -> Dict[Union[Hashable, str, Any], Any]:
        """Get the formatoptions for a new plot.
        
        This method updates the coordinates combo boxes with the 
        x- and y-coordinate of the variable.
        
        Parameters
        ----------
        var: xarray.Variable
            The variable in the base dataset

        Returns
        -------
        dict
            A mapping from formatoption or dimension to the corresponding value
            for the plotmethod.
        """
        ret = super().init_dims(var)

        dims: Dict[Hashable, Union[int, slice]] = {}
        xdim = ydim = None

        if self.combo_xdim.currentIndex():
            xdim = self.combo_xdim.currentText()
            if xdim in var.dims:
                dims[xdim] = slice(None)

        if self.combo_ydim.currentIndex():
            ydim = self.combo_ydim.currentText()
            if ydim in var.dims:
                dims[ydim] = slice(None)

        if dims:
            missing = [dim for dim in var.dims if dim not in dims]
            for dim in missing:
                dims[dim] = 0
            if len(dims) == 1 and xdim != ydim:
                if xdim is None:
                    xdim = missing[-1]
                else:
                    ydim = missing[-1]
                dims[missing[-1]] = slice(None)  # keep the last dimension
            ret['dims'] = dims


        if self.combo_xcoord.currentIndex():
            xcoord = self.combo_xcoord.currentText()
            ret['decoder'] = {'x': {xcoord}}
        if self.combo_ycoord.currentIndex():
            ycoord = self.combo_ycoord.currentText()
            ret.setdefault('decoder', {})
            ret['decoder']['y'] = {ycoord}

        if (xdim is not None and xdim in var.dims and
                ydim is not None and ydim in var.dims):
            ret['transpose'] = var.dims.index(xdim) < var.dims.index(ydim)

        return ret

    def valid_variables(self, ds: Dataset) -> List[Hashable]:
        """Get a list of variables that can be visualized with this plotmethod.

        Parameters
        ----------
        ds: xarray.Dataset
            The dataset to use

        Returns
        -------
        list of str
            List of variable names to plot
        """
        valid = super().valid_variables(ds)
        if (not any(combo.count() for combo in self.coord_combos) or
                not any(combo.currentIndex() for combo in self.coord_combos)):
            return valid
        if self.combo_xdim.currentIndex():
            xdim = self.combo_xdim.currentText()
            valid = [v for v in valid if xdim in ds[v].dims]
        if self.combo_ydim.currentIndex():
            ydim = self.combo_xdim.currentText()
            valid = [v for v in valid if ydim in ds[v].dims]
        if self.combo_xcoord.currentIndex():
            xc_dims = set(ds[self.combo_xcoord.currentText()].dims)
            valid = [v for v in valid
                     if xc_dims.intersection(ds[v].dims)]
        if self.combo_ycoord.currentIndex():
            yc_dims = set(ds[self.combo_ycoord.currentText()].dims)
            valid = [v for v in valid
                     if yc_dims.intersection(ds[v].dims)]
        return valid

    @property
    def coord_combos(self) -> List[QtWidgets.QComboBox]:
        """Get the combo boxes for x- and y-dimension and -coordinates."""
        return [self.combo_xdim, self.combo_ydim, self.combo_xcoord,
                self.combo_ycoord]

    @contextlib.contextmanager
    def block_combos(self) -> Iterator[None]:
        """Temporarilly block any signal of the :attr:`coord_combos`."""
        for combo in self.coord_combos:
            combo.blockSignals(True)
        yield
        for combo in self.coord_combos:
            combo.blockSignals(False)

    def setEnabled(self, b: bool) -> None:
        """Enable or disable the projection and color buttons.
        
        Parameters
        ----------
        b: bool
            If True, enable the buttons, else disable.
        """
        self.btn_proj_settings.setEnabled(b)
        self.proj_settings_action.setEnabled(b)
        self.btn_datagrid.setEnabled(b)
        self.color_settings_action.setEnabled(b)
        self.btn_cmap_settings.setEnabled(b)
        self.btn_labels.setEnabled(b)

    def set_cmap(self, cmap: str) -> None:
        """Update the plotter with the given colormap.

        Parameters
        ----------
        cmap: str
            The colormap name.
        """
        plotter = self.plotter
        if plotter and 'cmap' in plotter:
            plotter.update(cmap=cmap)

    def toggle_datagrid(self) -> None:
        """Toggle the visibility of the grid cell boundaries."""
        if self.plotter:
            if self.btn_datagrid.isChecked():
                self.plotter.update(datagrid='k--')
            else:
                self.plotter.update(datagrid=None)

    def edit_labels(self) -> None:
        """Open the dialog to edit the text labels in the plot."""
        dialogs.LabelDialog.update_project(
            self.sp, 'figtitle', 'title', 'clabel')

    def edit_color_settings(self) -> None:
        """Open the dialog to edit the color settings."""
        if self.sp and self.plotter:
            dialogs.CmapDialog.update_project(self.sp)
            if isinstance(self.plotter.cmap.value, str):
                self.btn_cmap.setText(self.plotter.cmap.value)
            else:
                self.btn_cmap.setText('Custom')

    def choose_next_projection(self) -> None:
        """Choose the next projection from the rcParams `projection` value."""
        select = False
        nprojections = len(rcParams['projections'])
        current = self.btn_proj.text()
        for i, proj in enumerate(cycle(rcParams['projections'])):
            if proj == current:
                select = True
            elif select or i == nprojections:
                break
        self.set_projection(proj)

    def set_projection(self, proj: str) -> None:
        """Update the projection of the plot with the given projection.
        
        Parameters
        ----------
        projection: str
            The projection name for the 
            :attr:`~psy_maps.plotters.FieldPlotter.projection` formatoption.
        """
        self.btn_proj.setText(proj)
        plotter = self.plotter
        if plotter and 'projection' in plotter:
            plotter.update(projection=proj)

    def edit_basemap_settings(self) -> None:
        """Open a dialog to edit the basemap and projection settings."""
        if self.plotter:
            dialogs.BasemapDialog.update_plotter(self.plotter)

    def get_fmts(
            self, var: DataArray,
            init: bool = False
        ) -> Dict[Union[Hashable, str, Any], Any]:
        """Get the formatoptions for a new plot.
        
        Parameters
        ----------
        var: xarray.Variable
            The variable in the base dataset
        init: bool
            If True, call the :meth:`init_dims` method to inject necessary
            formatoptions and dimensions for the initialization.

        Returns
        -------
        dict
            A mapping from formatoption or dimension to the corresponding value
            for the plotmethod.
        """
        fmts: Dict[Union[Hashable, str, Any], Any] = {}

        fmts['cmap'] = self.btn_cmap.text()

        if 'projection' in self.formatoptions:
            fmts['projection'] = self.btn_proj.text()

        if 'time' in var.dims:
            fmts['title'] = '%(time)s'

        if 'long_name' in var.attrs:
            fmts['clabel'] = '%(long_name)s'
        else:
            fmts['clabel'] = '%(name)s'
        if 'units' in var.attrs:
            fmts['clabel'] += ' %(units)s'

        if init:
            fmts.update(self.init_dims(var))

        return fmts

    def refresh(self, ds: Optional[Dataset]) -> None:
        """Refresh this widget from the given dataset."""
        self.setEnabled(bool(self.sp))

        auto = 'Set automatically'

        self.refresh_from_sp()

        with self.block_combos():

            if ds is None:
                ds = xr.Dataset()

            current_dims = set(map(
                self.combo_xdim.itemText, range(1, self.combo_xdim.count())))
            ds_dims = list(
                map(str, (dim for dim, n in ds.dims.items() if n > 1)))
            if current_dims != set(ds_dims):
                self.combo_xdim.clear()
                self.combo_ydim.clear()
                self.combo_xdim.addItems([auto] + ds_dims)
                self.combo_ydim.addItems([auto] + ds_dims)

            current_coords = set(map(
                self.combo_xcoord.itemText, range(1, self.combo_xcoord.count())))
            ds_coords = list(
                map(str, (c for c, arr in ds.coords.items() if arr.ndim)))
            if current_coords != set(ds_coords):
                self.combo_xcoord.clear()
                self.combo_ycoord.clear()
                self.combo_xcoord.addItems([auto] + ds_coords)
                self.combo_ycoord.addItems([auto] + ds_coords)

            enable_combos = not bool(self.sp)

            if not enable_combos and self.combo_xdim.isEnabled():
                self.reset_combos = [combo.currentIndex() == 0
                                    for combo in self.coord_combos]
            elif enable_combos and not self.combo_xdim.isEnabled():
                for reset, combo in zip(self.reset_combos, self.coord_combos):
                    if reset:
                        combo.setCurrentIndex(0)
                self.reset_combos = [False] * len(self.coord_combos)

            for combo in self.coord_combos:
                combo.setEnabled(enable_combos)

            if not enable_combos:
                data = self.data
                xdim = str(data.psy.get_dim('x'))
                ydim = str(data.psy.get_dim('y'))
                self.combo_xdim.setCurrentText(xdim)
                self.combo_ydim.setCurrentText(ydim)
                xcoord = data.psy.get_coord('x')
                xcoord = xcoord.name if xcoord is not None else xdim
                ycoord = data.psy.get_coord('y')
                ycoord = ycoord.name if ycoord is not None else ydim

                self.combo_xcoord.setCurrentText(xcoord)
                self.combo_ycoord.setCurrentText(ycoord)

    def refresh_from_sp(self) -> None:
        """Refresh this widget from the plotters project."""
        plotter = self.plotter
        if plotter:
            if isinstance(plotter.projection.value, str):
                self.btn_proj.setText(plotter.projection.value)
            else:
                self.btn_proj.setText('Custom')
            if isinstance(plotter.cmap.value, str):
                self.btn_cmap.setText(plotter.cmap.value)
            else:
                self.btn_cmap.setText('Custom')

    def transform(self, x: float, y: float) -> Tuple[float, float]:
        """Transform coordinates of a point to the plots projection.

        Parameters
        ----------
        x: float
            The x-coordinate in axes coordinates
        y: float
            The y-coordinate in axes coordinates

        Returns
        -------
        float
            The transformed x-coordinate `x`
        float
            The transformed y-coordinate `y`
        """
        import cartopy.crs as ccrs
        import numpy as np
        plotter = self.plotter
        if not plotter:
            raise ValueError(
                "Cannot transform the coordinates as no plot is shown "
                "currently!") 
        x, y = plotter.transform.projection.transform_point(
            x, y, plotter.ax.projection)
        # shift if necessary
        if isinstance(plotter.transform.projection, ccrs.PlateCarree):
            coord = plotter.plot.xcoord
            if coord.min() >= 0 and x < 0:
                x -= 360
            elif coord.max() <= 180 and x > 180:
                x -= 360
            if 'rad' in coord.attrs.get('units', '').lower():
                x = np.deg2rad(x)
                y = np.deg2rad(y)
        return x, y

    def get_slice(
            self, x: float, y: float
        ) -> Optional[Dict[Hashable, Union[int, slice]]]:
        """Get the slice for the selected coordinates.

        This method is called when the user clicks on the coordinates in the 
        plot.

        See Also
        --------
        psy_view.ds_widget.DatasetWidget.display_line

        Notes
        -----
        This is reimplemented in the :class:`MapPlotWidget`.
        """
        import numpy as np
        data = self.data.psy.base.psy[self.data.name]
        x, y = self.transform(x, y)
        plotter = self.plotter

        if not plotter:
            raise ValueError(
                "Cannot transform the coordinates as no plot is shown "
                "currently!")
                
        fmto = plotter.plot

        xcoord = fmto.xcoord
        ycoord = fmto.ycoord
        if fmto.decoder.is_unstructured(fmto.raw_data) or xcoord.ndim == 2:
            xy = xcoord.values.ravel() + 1j * ycoord.values.ravel()
            dist = np.abs(xy - (x + 1j * y))
            imin = np.nanargmin(dist)
            if xcoord.ndim == 2:
                ncols = data.shape[-2]
                return dict(zip(data.dims[-2:],
                                [imin // ncols, imin % ncols]))
            else:
                return {data.dims[-1]: imin}
        else:
            ix: int = xcoord.indexes[xcoord.name].get_loc(x, method='nearest')
            iy: int = ycoord.indexes[ycoord.name].get_loc(y, method='nearest')
            return dict(zip(data.dims[-2:], [iy, ix]))


class Plot2DWidget(MapPlotWidget):
    """A widget to control the plot2d plotmethod."""

    plotmethod = 'plot2d'

    def setup_projection_buttons(self) -> None:
        """Reimplemented to only show the datagrid button."""
        self.btn_datagrid = utils.add_pushbutton(
            "Cells", self.toggle_datagrid,
            "Show the grid cell boundaries", self.formatoptions_box)
        self.btn_datagrid.setCheckable(True)

    def setEnabled(self, b: bool) -> None:
        """Enable or disable the datagrid and color buttons.
        
        Parameters
        ----------
        b: bool
            If True, enable the buttons, else disable.
        """
        self.btn_datagrid.setEnabled(b)
        self.btn_cmap_settings.setEnabled(b)
        self.btn_labels.setEnabled(b)

    def edit_labels(self) -> None:
        """Open the dialog to edit the text labels in the plot."""
        dialogs.LabelDialog.update_project(
            self.sp, 'figtitle', 'title', 'clabel', 'xlabel', 'ylabel')

    def transform(self, x: float, y: float) -> Tuple[float, float]:
        """Reimplemented to not transform the coordinates."""
        return x, y

    def refresh_from_sp(self) -> None:
        """Refresh this widget from the plotters project."""
        plotter = self.plotter
        if plotter:
            if isinstance(plotter.cmap.value, str):
                self.btn_cmap.setText(plotter.cmap.value)


class LinePlotWidget(PlotMethodWidget):
    """A widget to control the lineplot plotmethod."""

    plotmethod = 'lineplot'

    def setup(self) -> None:
        """Set up widget during initialization."""
        self.layout = self.formatoptions_box = QtWidgets.QHBoxLayout()

        self.formatoptions_box.addWidget(QtWidgets.QLabel('x-Dimension:'))
        self.combo_dims = QtWidgets.QComboBox()
        self.combo_dims.setEditable(False)
        self.combo_dims.currentIndexChanged.connect(self.trigger_reset)
        self.formatoptions_box.addWidget(self.combo_dims)

        self.combo_lines = QtWidgets.QComboBox()
        self.combo_lines.setEditable(False)
        self.formatoptions_box.addWidget(self.combo_lines)
        self.combo_lines.setSizeAdjustPolicy(
            QtWidgets.QComboBox.AdjustToContents)
        self.combo_lines.currentIndexChanged.connect(self.trigger_refresh)
        self.formatoptions_box.addStretch(0)

        self.btn_add = utils.add_pushbutton(
            QtGui.QIcon(get_psy_icon('plus')), lambda: self.add_line(),
            "Add a line to the plot", self.formatoptions_box, icon=True)
        self.btn_del = utils.add_pushbutton(
            QtGui.QIcon(get_psy_icon('minus')), self.remove_line,
            "Add a line to the plot", self.formatoptions_box, icon=True)

        self.btn_labels = utils.add_pushbutton(
            "Labels", self.edit_labels,
            "Edit title, x-label, legendlabels, etc.", self.formatoptions_box)

    @property
    def xdim(self) -> str:
        """Get the x-dimension for the plot."""
        return self.combo_dims.currentText()

    @xdim.setter
    def xdim(self, xdim: Hashable) -> None:
        if xdim != self.combo_dims.currentText():
            self.combo_dims.setCurrentText(str(xdim))

    @property
    def data(self) -> DataArray:
        """The first array in the list."""
        data = super().data
        if len(data) - 1 < self.combo_lines.currentIndex():
            return data[0]
        else:
            return data[self.combo_lines.currentIndex()]

    def add_line(self, name: Hashable = None, **sl) -> None:
        """Add a line to the plot.

        This method adds a new line for the plot specified by the given 
        `name` of the variable and the slices.

        Parameters
        ----------
        name: str
            The variable name to display
        ``**sl``
            The slices to turn the `name` variable into a 1D-array.
        """
        if not self.sp:
            raise ValueError("No plot has yet been initialized!")
        ds = self.data.psy.base
        xdim = self.xdim
        if name is None:
            name, ok = QtWidgets.QInputDialog.getItem(
                self, 'New line', 'Select a variable',
                [key for key, var in ds.items()
                 if xdim in var.dims])
            if not ok:
                return
        arr = ds.psy[name]
        for key, val in self.data.psy.idims.items():
            if key in arr.dims:
                sl.setdefault(key, val)
        for dim in arr.dims:
            if dim != xdim:
                sl.setdefault(dim, 0)
        self.sp[0].append(arr.psy[sl], new_name=True)
        item = self.item_texts[-1]
        self.sp.update(replot=True)
        self.combo_lines.addItem(item)
        self.combo_lines.setCurrentText(item)
        self.trigger_refresh()

    def remove_line(self) -> None:
        """Remove the current line from the plot."""
        if not self.sp:
            raise ValueError(
                "No plot has yet been initialized, so I cannot remove any line!"
            )
        i = self.combo_lines.currentIndex()
        self.sp[0].pop(i)
        self.sp.update(replot=True)
        self.combo_lines.setCurrentText(self.item_texts[i - 1 if i else 0])
        self.changed.emit(self.plotmethod)

    @property
    def item_texts(self) -> List[str]:
        """Get the labels for the individual lines."""
        if not self.sp:
            return []
        return [f'Line {i}: {arr.psy._short_info()}'
                for i, arr in enumerate(self.sp[0])]

    def init_dims(
            self, var: DataArray
        ) -> Dict[Union[Hashable, str, Any], Any]:
        """Get the formatoptions for a new plot.
        
        Parameters
        ----------
        var: xarray.Variable
            The variable in the base dataset

        Returns
        -------
        dict
            A mapping from formatoption or dimension to the corresponding value
            for the plotmethod.
        
        """
        ret: Dict[Union[Hashable, str, Any], Any] = {}
        xdim: Union[None, Hashable, str] = self.xdim or next(
            (d for d in var.dims if var[d].size > 1), None  # type: ignore
        )
        if self.array_info:
            arr_names = {}
            for arrname, d in self.array_info.items():
                if arrname != 'attrs':
                    dims = d['dims'].copy()
                    if xdim in dims:
                        for dim, sl in dims.items():
                            if not isinstance(sl, int):
                                dims[dim] = 0
                        dims[xdim] = slice(None)
                    dims['name'] = d['name']
                    arr_names[arrname] = dims
            ret['arr_names'] = arr_names
            del self.array_info
        else:
            if xdim not in var.dims:
                xdim = next((d for d in var.dims if var[d].size > 1), None)
            if xdim is None:
                raise ValueError(
                    f"Cannot plot variable {var.name} with size smaller than "
                    "2")
            ret[xdim] = slice(None)
            for d in var.dims:
                if d != xdim:
                    ret[d] = 0
        return ret

    def edit_labels(self) -> None:
        """Open the dialog to edit the text labels in the plot."""
        dialogs.LabelDialog.update_project(
            self.sp, 'figtitle', 'title', 'xlabel', 'ylabel', 'legendlabels')

    @contextlib.contextmanager
    def block_combos(self) -> Iterator[None]:
        """Temporarilly block any signal of the combo boxes."""
        self.combo_dims.blockSignals(True)
        self.combo_lines.blockSignals(True)
        yield
        self.combo_dims.blockSignals(False)
        self.combo_lines.blockSignals(False)

    def valid_variables(self, ds: Dataset) -> List[Hashable]:
        """Get a list of variables that can be visualized with this plotmethod.

        Parameters
        ----------
        ds: xarray.Dataset
            The dataset to use

        Returns
        -------
        list of str
            List of variable names to plot
        """
        valid = list(ds)
        if not self.sp or len(self.sp[0]) < 2:
            return valid
        else:
            current_dim = self.combo_dims.currentText()
            return [v for v in valid if current_dim in ds[v].dims]

    def refresh(self, ds) -> None:
        """Refresh this widget from the given dataset."""
        if self.sp:
            with self.block_combos():
                self.combo_dims.clear()
                all_dims = list(chain.from_iterable(
                    [[d for i, d in enumerate(a.dims) if a.shape[i] > 1]
                     for a in arr.psy.iter_base_variables]
                    for arr in self.sp[0]))
                intersection = set(all_dims[0])
                for dims in all_dims[1:]:
                    intersection.intersection_update(dims)
                new_dims = list(
                    filter(lambda d: d in intersection,
                           unique_everseen(chain.from_iterable(all_dims))))

                self.combo_dims.addItems(new_dims)
                self.combo_dims.setCurrentIndex(
                    new_dims.index(self.data.dims[-1]))

                # fill lines combo
                current = self.combo_lines.currentIndex()
                self.combo_lines.clear()
                descriptions = self.item_texts
                short_descs = [textwrap.shorten(s, 50) for s in descriptions]
                self.combo_lines.addItems(short_descs)
                for i, desc in enumerate(descriptions):
                    self.combo_lines.setItemData(i, desc, QtCore.Qt.ToolTipRole)
                if current < len(descriptions):
                    self.combo_lines.setCurrentText(short_descs[current])
        else:
            with self.block_combos():
                self.combo_dims.clear()
                self.combo_lines.clear()
        self.btn_add.setEnabled(bool(self.sp))
        self.btn_del.setEnabled(bool(self.sp) and len(self.sp[0]) > 1)



# -*- coding: utf-8 -*-
"""Dataset widget to display the contents of a dataset

Copyright (C) 2020 Philipp S. Sommer

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see https://www.gnu.org/licenses/."""
from __future__ import annotations

import os.path as osp
import os

from typing import (
    List,
    TYPE_CHECKING,
    Optional,
    Union,
    Dict,
    Iterator,
    Type,
    Any,
    Callable,
    Hashable,
    Tuple,
)

import contextlib

import yaml
from PyQt5 import QtWidgets, QtGui
from PyQt5.QtCore import Qt  # pylint: disable=no-name-in-module
import psy_view.utils as utils
from psyplot_gui.content_widget import (
    DatasetTree, DatasetTreeItem, escape_html)
from psyplot_gui.common import (
    DockMixin, get_icon as get_psy_icon, PyErrorMessage)
import psyplot.data as psyd
from psy_view.rcsetup import rcParams
from psy_view.plotmethods import (
    PlotMethodWidget,
    MapPlotWidget,
    Plot2DWidget,
    LinePlotWidget,
)
from psyplot.config.rcsetup import get_configdir

from matplotlib.animation import FuncAnimation

if TYPE_CHECKING:
    from xarray import DataArray, Dataset
    from psyplot.project import PlotterInterface, Project
    from psyplot.plotter import Plotter
    from matplotlib.figure import Figure
    from matplotlib.backend_bases import MouseEvent
    from psyplot_gui.main import MainWindow

NOTSET_T = Type[object]
NOTSET: NOTSET_T = object


def get_dims_to_iterate(arr: DataArray) -> List[str]:
    """Get the dimensions of an array to iterate over

    This function takes a data array and returns the dimension in the base
    dataset that one can interator over.

    Parameters
    ----------
    arr: xarray.DataArray
        The data array to iterate over

    Returns
    -------
    list of strings
        The dimension strings
    """
    base_var = next(arr.psy.iter_base_variables)
    return [dim for dim, size in zip(base_var.dims, base_var.shape)
            if size > 1 and arr[dim].ndim == 0]

TOO_MANY_FIGURES_WARNING = """
Multiple figures are open but you specified only {} filenames: {}.<br>

Saving the figures will cause that not all images are saved! We recommend to
export to a single PDF (that then includes multiple pages), or modify your
filename with strings like

<ul>
<li> <code>%i</code> for a continuous counter of the images</li>
<li><code>%(name)s</code> for variable names</li>
<li>or other netCDF attributes (see the
  <a href="https://psyplot.readthedocs.io/en/latest/api/psyplot.project.html#psyplot.project.Project.export">
  examples of exporting psyplot projects</a>)</li>
</ul>

Shall I continue anyway and save the figures?
"""


class DatasetWidget(QtWidgets.QSplitter):
    """A widget to control the visualization of the variables in a dataset"""

    #: The title of the widget
    title: str = 'psy-view Plot Control'

    #: Display the dock widget at the right side of the GUI
    dock_position = Qt.RightDockWidgetArea

    _animating: bool = False

    _ani: Optional[FuncAnimation] = None

    _init_step: int = 0

    #: A :class:`PyQt5.QtWidgets.QGroupBox` that contains the variable buttons
    variable_frame: Optional[QtWidgets.QGroupBox] = None

    _new_plot: bool = False

    _preset: Optional[Union[str, Dict]] = None

    #: Attributes to use in the dataset tree
    ds_attr_columns: List[str] = ['long_name', 'dims', 'shape']

    def __init__(self, ds: Optional[Dataset] = None, *args, **kwargs) -> None:
        """
        Parameters
        ----------
        ds: xarray.Dataset
            A dataset to visualize with this widget
        """
        super().__init__(*args, **kwargs)

        self._ds_nums: Dict[int, Dataset] = {}

        self.setChildrenCollapsible(False)

        self.ds: Optional[Dataset] = ds

        self.setOrientation(Qt.Vertical)

        self.error_msg = PyErrorMessage(self)

        # first row: dataset name
        self.open_box = QtWidgets.QHBoxLayout()
        self.lbl_ds = QtWidgets.QLineEdit()
        self.open_box.addWidget(self.lbl_ds)
        self.btn_open = utils.add_pushbutton(
            get_psy_icon('run_arrow.png'), lambda: self.set_dataset(),
            "Select and open a netCDF dataset", self.open_box, icon=True)
        self.open_widget = QtWidgets.QWidget()
        self.open_widget.setLayout(self.open_box)
        self.addWidget(self.open_widget)


        # second row: dataset representation
        self.setup_ds_tree()
        if ds is not None:
            self.add_ds_item()

        self.ds_tree.itemExpanded.connect(self.change_ds)
        self.ds_tree.itemExpanded.connect(self.load_variable_desc)

        self.addWidget(self.ds_tree)

        # third row, navigation
        self.navigation_box = QtWidgets.QHBoxLayout()

        # -- animate backwards button
        self.btn_animate_backward = utils.add_pushbutton(
            "◀◀", lambda: self.animate_backward(),
            "Animate the time dimension backwards", self.navigation_box)
        self.btn_animate_backward.setCheckable(True)

        # -- go to previous button
        self.btn_prev = utils.add_pushbutton(
            '◀', self.go_to_previous_step,
            "Go to previous time step", self.navigation_box)

        # -- dimension menu for animation
        self.combo_dims = QtWidgets.QComboBox()
        self.navigation_box.addWidget(self.combo_dims)

        # -- go to next button
        self.btn_next = utils.add_pushbutton(
            '▶', self.go_to_next_step,
            "Go to next time step", self.navigation_box)

        # -- animate forward button
        self.btn_animate_forward = utils.add_pushbutton(
            "▶▶", lambda: self.animate_forward(),
            "Animate the time dimension", self.navigation_box)
        self.btn_animate_forward.setCheckable(True)

        # -- interval slider
        self.sl_interval = QtWidgets.QSlider(Qt.Horizontal)
        self.sl_interval.setMinimum(40)  # 24 fps
        self.sl_interval.setMaximum(10000)
        self.sl_interval.setSingleStep(50)
        self.sl_interval.setPageStep(500)
        self.sl_interval.setValue(500)
        self.sl_interval.valueChanged.connect(self.reset_timer_interval)
        self.navigation_box.addWidget(self.sl_interval)

        # -- interval label
        self.lbl_interval = QtWidgets.QLabel('500 ms')
        self.navigation_box.addWidget(self.lbl_interval)

        # --- export/import menus
        self.export_box = QtWidgets.QHBoxLayout()

        # -- Export button
        self.btn_export = QtWidgets.QToolButton()
        self.btn_export.setText('Export')
        self.btn_export.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.btn_export.setMenu(self.setup_export_menu())
        self.btn_export.setEnabled(False)
        self.export_box.addWidget(self.btn_export)

        # --- Presets button
        self.frm_preset = QtWidgets.QFrame()
        self.frm_preset.setFrameStyle(QtWidgets.QFrame.StyledPanel)
        hbox = QtWidgets.QHBoxLayout(self.frm_preset)

        self.btn_preset = QtWidgets.QToolButton()
        self.btn_preset.setText('Preset')
        self.btn_preset.setPopupMode(QtWidgets.QToolButton.InstantPopup)
        self.btn_preset.setMenu(self.setup_preset_menu())
        hbox.addWidget(self.btn_preset)

        # --- presets label
        self.lbl_preset = QtWidgets.QLabel('')
        self.lbl_preset.setVisible(False)
        hbox.addWidget(self.lbl_preset)

        # --- unset preset button
        self.btn_unset_preset = utils.add_pushbutton(
            get_psy_icon('invalid.png'), self.unset_preset,
            "Unset the current preset", hbox, icon=True)
        self.btn_unset_preset.setVisible(False)

        self.export_box.addWidget(self.frm_preset)

        self.export_box.addStretch(0)

        vbox = QtWidgets.QVBoxLayout()
        vbox.addLayout(self.navigation_box)
        vbox.addLayout(self.export_box)

        self.addLayout(vbox)

        # fourth row: array selector

        self.array_frame = QtWidgets.QGroupBox('Current plot')
        hbox = QtWidgets.QHBoxLayout()

        self.combo_array = QtWidgets.QComboBox()
        self.combo_array.setEditable(False)
        self.combo_array.currentIndexChanged.connect(lambda: self.refresh())
        self.combo_array.currentIndexChanged.connect(self.show_current_figure)
        hbox.addWidget(self.combo_array)

        self.btn_add = utils.add_pushbutton(
            QtGui.QIcon(get_psy_icon('plus')), self.new_plot,
            "Create a new plot", hbox, icon=True)
        self.btn_add.setEnabled(ds is not None)
        self.btn_del = utils.add_pushbutton(
            QtGui.QIcon(get_psy_icon('minus')), self.close_current_plot,
            "Remove the current plot", hbox, icon=True)
        self.btn_del.setEnabled(False)

        hbox.addWidget(self.btn_add)
        hbox.addWidget(self.btn_del)
        self.array_frame.setLayout(hbox)
        self.addWidget(self.array_frame)

        # fifth row: plot interface
        self.plot_tabs = QtWidgets.QTabWidget()
        self.setup_plot_tabs()
        self.plot_tabs.currentChanged.connect(self.switch_tab)

        self.addWidget(self.plot_tabs)

        # sixth row: variables
        self.variable_scroll = QtWidgets.QScrollArea()
        self.variable_scroll.setWidgetResizable(True)
        self.setup_variable_buttons()
        self.addWidget(self.variable_scroll)

        # seventh row: dimensions
        self.dimension_table = QtWidgets.QTableWidget()
        self.addWidget(self.dimension_table)

        self.disable_navigation()

        if self.ds is not None:
            self.refresh()

        self.cids: Dict[str, int] = {}

    def setup_ds_tree(self) -> None:
        """Setup the number of columns and the header of the dataset tree."""
        self.ds_tree = tree = QtWidgets.QTreeWidget()
        tree.setColumnCount(len(self.ds_attr_columns) + 1)
        tree.setHeaderLabels([''] + self.ds_attr_columns)

    def showEvent(self, event):
        ret = super().showEvent(event)
        current_size = self.size()
        current_sizes = self.sizes()
        new_sizes = list(current_sizes)
        itree = self.indexOf(self.ds_tree)
        itable = self.indexOf(self.dimension_table)
        diff = 0
        if current_sizes[itree] < 400:
            diff += 400 - current_sizes[itree]
            current_sizes[itree] = 400
        if current_sizes[itable] < 300:
            diff += 300 - current_sizes[itable]
            current_sizes[itable] = 300
        if diff:
            self.resize(current_size.width(), current_size.height() + diff)
            self.setSizes(current_sizes)
        return ret

    def close_current_plot(self) -> None:
        """Close the figure of the current variable."""
        self.variable_buttons[self.variable].click()

    def excepthook(self, type, value, traceback) -> None:
        """A method to replace the sys.excepthook"""
        self.error_msg.excepthook(type, value, traceback)

    @property
    def arr_name(self) -> Optional[str]:
        """Get the name of the array of the current plot (if there is one)."""
        if not self.combo_array.count():
            return None
        else:
            return self.combo_array.currentText().split(':')[0]

    def change_ds(self, ds_item: DatasetTreeItem) -> None:
        """Change the current dataset to another one.

        Parameters
        ----------
        ds_item: psyplot_gui.content_widget.DatasetTreeItem
            The item in the tree of the new dataset to use
        """
        ds_items = self.ds_items
        if ds_item in ds_items:
            with self.block_tree():
                self.ds = ds_item.ds()
                self.expand_ds_item(ds_item)
                self.setup_variable_buttons()
                self.change_combo_array()
                self.refresh(reset_combo=False)

    def expand_ds_item(self, ds_item: DatasetTreeItem) -> None:
        """Expand an item of a dataset.

        Parameters
        ----------
        ds_item: DatasetTreeItem
            The item to expand
        """
        tree = self.ds_tree

        tree.collapseAll()

        tree.expandItem(ds_item)

        ds = ds_item.ds()
        if len(ds) <= 10:
            tree.expandItem(ds_item.child(0))
        if len(ds.coords) <= 10:
            tree.expandItem(ds_item.child(1))
        if len(ds.attrs) <= 10:
            tree.expandItem(ds_item.attrs)

    def _open_dataset(self) -> Optional[Dataset]:
        """Open a dialog to open a new dataset from disk.

        Returns
        -------
        xarray.Dataset or None
            The :class:`xarray.Dataset` of the selected file, or None if the
            user aborted the dialog.
        """
        current = self.lbl_ds.text()
        if not current or not osp.exists(current):
            current = os.getcwd()
        fname, ok = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Open dataset', current,
            'NetCDF files (*.nc *.nc4);;'
            'Shape files (*.shp);;'
            'All files (*)'
            )
        if not ok:
            return None
        ds = psyd.open_dataset(fname)
        return ds

    @contextlib.contextmanager
    def block_tree(self) -> Iterator[None]:
        """Block all signals of a tree temporarily.

        Use this via::

            with self.block_tree():
                do_something
        """
        self.ds_tree.blockSignals(True)
        yield
        self.ds_tree.blockSignals(False)

    def set_dataset(self, ds: Optional[Dataset] = None) -> None:
        """Ask for a file name and open the dataset."""
        if ds is None:
            ds = self._open_dataset()
        if ds is None:
            return
        self.ds = ds
        with self.block_tree():
            self.add_ds_item()
            self.setup_variable_buttons()
            self.btn_add.setEnabled(True)
            self.btn_del.setEnabled(True)

    def add_ds_item(self) -> None:
        """Add a new :class:`DatasetTreeItem` for the current :attr:`ds`."""
        ds: Dataset = self.ds  # type: ignore
        tree = self.ds_tree
        ds_item = DatasetTreeItem(ds, self.ds_attr_columns, 0)
        fname = psyd.get_filename_ds(ds, False)[0]
        if fname is not None:
            self.lbl_ds.setText(fname)
            fname = osp.basename(fname)
        else:
            self.lbl_ds.setText('')
            fname = ''
        ds_item.setText(0, fname)
        tree.addTopLevelItem(ds_item)

        self.expand_ds_item(ds_item)

        tree.resizeColumnToContents(0)

        if ds.psy.num not in self.open_datasets:
            # make sure we do not loose track of open datasets
            self._ds_nums[ds.psy.num] = ds

    @property
    def open_datasets(self) -> Dict[int, Dataset]:
        """Get a mapping from path to dataset number of the open datasets."""
        return self._ds_nums

    @property
    def ds_items(self) -> List[DatasetTreeItem]:
        """Get the :class:`DatasetTreeItems` for the open datasets."""
        tree = self.ds_tree
        return list(map(tree.topLevelItem, range(tree.topLevelItemCount())))

    @property
    def ds_item(self) -> Optional[DatasetTreeItem]:
        """Get the current dataset item (if there is one)."""
        ds = self.ds
        for item in self.ds_items:
            if item.ds() is ds:
                return item
        return None

    def expand_current_variable(
            self, variable: Optional[Union[Any, Hashable]] = None) -> None:
        """Expand the item in the dataset tree of variable.

        Parameters
        ----------
        variable: str
            The name of the variable to expand. If None, the current variable is
            used.
        """
        tree = self.ds_tree
        top: DatasetTreeItem = self.ds_item  # type: ignore
        tree.expandItem(top)
        tree.expandItem(top.child(0))
        if variable is None:
            variable: str = self.variable  # type: ignore
        for var_item in map(top.child(0).child,
                            range(top.child(0).childCount())):
            if var_item.text(0) == variable:
                tree.expandItem(var_item)
            else:
                tree.collapseItem(var_item)

    def setup_variable_buttons(self, ncols: int = 4) -> None:
        """Setup the variable buttons for the current dataset."""
        variable_frame = QtWidgets.QGroupBox('Variables')

        self.variable_scroll.setWidget(variable_frame)
        self.variable_frame = variable_frame
        self.variable_layout = QtWidgets.QGridLayout(self.variable_frame)
        self.variable_buttons = {}

        ds = self.ds

        if ds is not None:

            for i, v in enumerate(ds):
                btn = utils.add_pushbutton(
                    v, self._draw_variable(v), f"Visualize variable {v}")
                btn.setCheckable(True)
                self.variable_buttons[v] = btn
                self.variable_layout.addWidget(btn, i // ncols, i % ncols)

            if len(ds):
                rows = len(ds) // ncols
                minrows = max(1, min(3, rows))
                self.variable_scroll.setMinimumHeight(
                        (minrows + 2) * btn.sizeHint().height())

    def load_variable_desc(self, item: QtWidgets.QTreeWidgetItem) -> None:
        """Load the description of the variable of a given tree item.

        Parameters
        ----------
        item: PyQt5.QtWidget.QTreeWidgetItem
            The item of the variable in the :attr:`ds_tree`. If this is not an
            item of a variable, nothing is done.
        """
        parent = item.parent()

        tree = self.ds_tree

        if parent is tree or parent is None or not (
                DatasetTree.is_variable(item) or DatasetTree.is_coord(item)):
            return

        if tree.isColumnHidden(1):
            tree.showColumn(1)
            tree.resizeColumnToContents(0)

        top = item
        while top.parent() and top.parent() is not self:
            top = top.parent()
        ds = top.ds()
        if ds is None:
            return
        desc = escape_html(str(ds.variables[item.text(0)]))
        item.setToolTip(0, '<pre>' + desc + '</pre>')

    def clear_table(self) -> None:
        """Clear the table that shows the available dimensions."""
        self.dimension_table.clear()
        self.dimension_table.setColumnCount(5)
        self.dimension_table.setHorizontalHeaderLabels(
            ['Type', 'First', 'Current', 'Last', 'Units'])
        self.dimension_table.setRowCount(0)

    def addLayout(self, layout: QtWidgets.QLayout) -> QtWidgets.QWidget:
        """Add a layout to the splitter.

        This convenience function creates a new QWidget that wraps the given
        layout and returns it.

        Parameters
        ----------
        layout: QtWidget.QLayout
            The layout to add

        Returns
        -------
        QtWidgets.QWidget
            The widget that wraps the given layout
        """
        widget = QtWidgets.QWidget()
        widget.setLayout(layout)
        self.addWidget(widget)
        return widget

    def go_to_previous_step(self) -> None:
        """Decrease the movie dimension to the previous step."""
        dim = self.combo_dims.currentText()
        self.increase_dim(dim, -1)()

    def go_to_next_step(self) -> None:
        """Increase the movie dimension to the next step."""
        dim = self.combo_dims.currentText()
        self.increase_dim(dim)()

    def animate_backward(self) -> None:
        """Start the current animation in backward direction, or stop it."""
        if self._animating:
            self.stop_animation()
            self.btn_animate_backward.setText('◀◀')
            self.enable_navigation()
        else:
            self._animate_forward = False
            self.btn_animate_backward.setText('■')
            self.disable_navigation(self.btn_animate_backward)
            self.start_animation()

    def animate_forward(self, nframes=None):
        """Start the current animation in forward direction, or stop it."""
        if self._animating:
            self.stop_animation()
            self.btn_animate_forward.setText('▶▶')
            self.enable_navigation()
        else:
            self._animate_forward = True
            self.btn_animate_forward.setText('■')
            self.disable_navigation(self.btn_animate_forward)
            self.start_animation(nframes)

    def setup_plot_tabs(self) -> None:
        """Setup the tabs of the various plot methods."""
        self.plot_tabs.addTab(MapPlotWidget(self.get_sp, self.ds),
                              'mapplot')
        self.plot_tabs.addTab(Plot2DWidget(self.get_sp, self.ds),
                              'plot2d')
        lineplot_widget = LinePlotWidget(self.get_sp, self.ds)
        self.plot_tabs.addTab(lineplot_widget, 'lineplot')

        for w in map(self.plot_tabs.widget, range(self.plot_tabs.count())):
            w.replot.connect(self.replot)
            w.reset.connect(self.reset)
            w.changed.connect(lambda: self.refresh())

    def replot(self, plotmethod: str) -> None:
        """Regenerate the plot of a given plotmethod, without closing it.

        Parameters
        ----------
        plotmethod: str
            The name of the plotmethod

        See Also
        --------
        reset: The same method, but closes the plot before genereting a new one.
        """
        self.plotmethod = plotmethod
        self.make_plot()
        self.refresh()

    def reset(self, plotmethod: str) -> None:
        """Close the plot of the given plotmethod and regenerate it.

        The same as :meth:`replot`, but closes the plot.

        Parameters
        ----------
        plotmethod: str
            The name of the plotmethod

        See Also
        --------
        reset: The same method, but closes the plot before genereting a new one.
        """
        self.plotmethod = plotmethod
        self.close_sp()
        self.make_plot()
        self.refresh()

    def disable_navigation(
            self, but: Optional[QtWidgets.QPushButton] = None
        ) -> None:
        """Disable the navigation buttons.

        This function disables all navigation buttons but the one you specify.

        Parameters
        ----------
        but: PyQt5.QtWidgets.QPushButton
            If not None, this button is not disabled.
        """
        for item in map(self.navigation_box.itemAt,
                        range(self.navigation_box.count())):
            w = item.widget()
            if w is not but and w is not self.sl_interval:
                w.setEnabled(False)

    def enable_navigation(self) -> None:
        """Enable all navigation buttons again."""
        for item in map(self.navigation_box.itemAt,
                        range(self.navigation_box.count())):
            w = item.widget()
            w.setEnabled(True)

    def disable_variables(self):
        """Disable all variable selection buttons."""
        for btn in self.variable_buttons.values():
            btn.setEnabled(False)

    def enable_variables(self):
        """Enable all variable selection buttons again."""
        valid_variables = self.plotmethod_widget.valid_variables(self.ds)
        for v, btn in self.variable_buttons.items():
            btn.setEnabled(v in valid_variables)

    def start_animation(self, nframes: Optional[int] = None):
        """Start the animation along the selected dimension.

        Parameters
        ----------
        nframes: int or None
            If not None, the number of frames to draw

        See Also
        --------
        animation_frames: The iterator to generate the frames
        """
        self._animating = True
        self._animation_frames = nframes
        self._starting_step = 1
        self.disable_variables()
        self.plot_tabs.setEnabled(False)
        if self.sp is not None:
            if self.animation is None or self.animation.event_source is None:
                self.animation = FuncAnimation(
                    self.fig, self.update_dims, frames=self.animation_frames(),
                    init_func=self.sp.draw, interval=self.sl_interval.value(),
                    repeat=False)
                # HACK: Make sure that the animation starts although the figure
                # is already shown
                self.animation._draw_frame(next(self.animation_frames()))
            else:
                self.animation.event_source.start()

    def reset_timer_interval(self, value: int) -> None:
        """Change the interval of the timer."""
        self.lbl_interval.setText('%i ms' % value)
        if self.animation is None or self.animation.event_source is None:
            pass
        else:
            self.animation.event_source.stop()
            self.animation._interval = value
            self.animation.event_source.interval = value
            self.animation.event_source.start()

    def stop_animation(self) -> None:
        """Stop the current animation."""
        self._animating = False
        if (self.animation is not None and
                self.animation.event_source is not None):
            self.animation.event_source.stop()
        self.plot_tabs.setEnabled(True)
        self.enable_variables()
        self.refresh()

    def animation_frames(self) -> Iterator[Dict[str, int]]:
        """Get the animation frames for the :attr:`combo_dims` dimension."""
        while self._animating and self._animation_frames is None or \
                self._animation_frames:
            if self._animation_frames is not None and not self._init_step:
                self._animation_frames -= 1
            dim = self.combo_dims.currentText()
            i = self.data.psy.idims[dim]
            imax = self.ds.dims[dim] - 1  # type: ignore
            if self._init_step:
                self._init_step -= 1
            elif self._starting_step:
                self._starting_step -= 1
            elif self._animate_forward:
                i += -i if i == imax else 1
            else:
                i += imax if i == 0 else -1
            yield {dim: i}

    def update_dims(self, dims: Dict[str, Any]):
        if self.sp is not None:
            self.sp.update(dims=dims)

    def _load_preset(self) -> None:
        """Open a file dialog and load the selected preset."""
        fname, ok = QtWidgets.QFileDialog.getOpenFileName(
            self, 'Load preset', osp.join(get_configdir(), 'presets'),
            'YAML files (*.yml *.yaml);;'
            'All files (*)')
        if ok:
            self.load_preset(fname)

    def load_preset(self, preset: Optional[Union[str, Dict[str, Any]]]):
        """Load a given preset from disk.
        
        Parameters
        ----------
        preset: str or dict
            The name or path to the preset, or a dictionary
        """
        self.preset = preset  # type: ignore
        if self.sp:
            loaded_preset: Dict[str, Any] = self.preset  # now that it's loaded
            if loaded_preset:
                self.sp.load_preset(loaded_preset)
                self.refresh()
        self.maybe_show_preset()

    @property
    def preset(self) -> Dict[str, Any]:
        """Get the currently loaded preset."""
        if self._preset is None:
            return {}
        import psyplot.project as psy
        preset = self._preset
        try:
            preset = psy.Project._load_preset(preset)
        except yaml.constructor.ConstructorError:
            answer = QtWidgets.QMessageBox.question(
                self, "Can I trust this?",
                f"Failed to load the preset at <i>{preset}</i> in safe mode. Can we "
                "trust this preset and load it in unsafe mode?")
            if answer == QtWidgets.QMessageBox.Yes:
                psyd.rcParams['presets.trusted'].append(
                    psy.Project._resolve_preset_path(preset))
                preset = psy.Project._load_preset(preset)
            else:
                preset = {}
        return preset  # type: ignore

    @preset.setter
    def preset(self, value: Optional[Union[str, Dict[str, Any]]]):
        self._preset = value


    def unset_preset(self) -> None:
        """Unset the current preset and do not use it anymore."""
        self.preset = None  # type: ignore
        self.maybe_show_preset()

    def maybe_show_preset(self) -> None:
        """Show the name of the current preset if one is selected."""
        if self._preset is not None and isinstance(self._preset, str):
            self.lbl_preset.setText('<i>' +
                osp.basename(osp.splitext(self._preset)[0]) + '</i>')
            self.lbl_preset.setVisible(True)
            self.btn_unset_preset.setVisible(True)
        elif self._preset is not None:
            self.lbl_preset.setText('<i>custom</i>')
            self.lbl_preset.setVisible(True)
            self.btn_unset_preset.setVisible(True)
        else:
            self.lbl_preset.setVisible(False)
            self.btn_unset_preset.setVisible(False)

    def save_current_preset(self) -> None:
        """Save the preset of the current plot to a file."""
        if self.sp is not None:
            preset_func = self.sp.save_preset
            self._save_preset(preset_func)

    def save_full_preset(self) -> None:
        """Save the preset of all open plots to a file."""
        sp = self._sp
        if sp is not None:
            return self._save_preset(sp.save_preset)
        return None

    def _save_preset(self, save_func: Callable[[str], Any]) -> None:
        """Save the preset to a file.

        Parameters
        ----------
        save_func: function
            The function that is called to save the preset. Must accept the
            path as an argument
        """
        fname, ok = QtWidgets.QFileDialog.getSaveFileName(
            self, 'Save preset', osp.join(get_configdir(), 'presets'),
            'YAML files (*.yml *.yaml);;'
            'All files (*)')
        if not ok:
            return None
        save_func(fname)

    def setup_preset_menu(self) -> QtWidgets.QMenu:
        """Set up the menu to select/load presets."""
        self.preset_menu = menu = QtWidgets.QMenu()
        self._save_preset_actions = []

        self._load_preset_action = menu.addAction(
            "Load preset", self._load_preset)
        self._unset_preset_action = menu.addAction(
            "Unset preset", self.unset_preset)

        menu.addSeparator()

        self._save_preset_actions.append(
            menu.addAction('Save format of current plot as preset',
                           self.save_current_preset))
        self._save_preset_actions.append(
            menu.addAction('Save format of all plots as preset',
                           self.save_full_preset))

        for action in self._save_preset_actions:
            action.setEnabled(False)

        return menu

    def setup_export_menu(self) -> QtWidgets.QMenu:
        """Set up the menu to export the current plot."""
        self.export_menu = menu = QtWidgets.QMenu()
        menu.addAction('image (PDF, PNG, etc.)', self.export_image)
        menu.addAction('all images (PDF, PNG, etc.)', self.export_all_images)
        menu.addAction('animation (GIF, MP4, etc.', self.export_animation)
        menu.addAction('psyplot project (.pkl file)', self.export_project)
        menu.addAction('psyplot project with data',
                       self.export_project_with_data)
        py_action = menu.addAction('python script (.py)', self.export_python)
        py_action.setEnabled(False)  # psyplot does not yet export to python
        return menu

    def export_image(self) -> None:
        """Ask for a filename and export the current plot to a file."""
        if self.sp is not None:
            fname, ok = QtWidgets.QFileDialog.getSaveFileName(
                self, "Export image", os.getcwd(),
                "Images (*.png *.pdf *.jpg *.svg)")
            if ok:
                self.sp.export(fname, **rcParams['savefig_kws'])

    def export_all_images(self) -> None:
        """Ask for a filename and export all plots to one (or more) files."""
        fname, ok = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export image", os.getcwd(),
            "Images (*.png *.pdf *.jpg *.svg)")
        if ok and self._sp:
            # test filenames
            if not osp.splitext(fname)[-1].lower() == '.pdf':
                fnames = [
                    sp.format_string(fname, False, i)
                    for i, sp in enumerate(self._sp.figs.values())]
                if len(fnames) != len(set(fnames)):
                    answer = QtWidgets.QMessageBox.question(
                        self, "Too many figures",
                        TOO_MANY_FIGURES_WARNING.format(
                            len(set(fnames)), ', '.join(set(fnames))))
                    if answer == QtWidgets.QMessageBox.No:
                        return
            self._sp.export(fname, **rcParams['savefig_kws'])

    def export_animation(self) -> None:
        """Ask for a filename and export the animation."""
        fname, ok = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export animation", os.getcwd(),
            "Movie (*.mp4 *.mov *.gif)")
        if ok:
            dim = self.combo_dims.currentText()
            nframes: int = self.ds.dims[dim]  # type: ignore

            self._init_step = 1
            self.animate_forward(nframes)
            if self.animation is not None:
                self.animation.save(
                    fname, **rcParams['animations.export_kws'],
                    fps=round(1000. / self.sl_interval.value()))
            self.animate_forward()
            self.animation = None

    def export_project(self) -> None:
        """Ask for a filename and export the psyplot project as .pkl file."""
        if self.sp is not None:
            fname, ok = QtWidgets.QFileDialog.getSaveFileName(
                self, "Export project", os.getcwd(),
                "Psyplot projects (*.pkl)")
            if ok:
                self.sp.save_project(fname)

    def export_project_with_data(self) -> None:
        """Ask for a filename export project (incl. data) as .pkl file.

        Same as :meth:`export_project`, but adds the data to the pickle dump.
        """
        if self.sp is not None:
            fname, ok = QtWidgets.QFileDialog.getSaveFileName(
                self, "Export project", os.getcwd(),
                "Psyplot projects (*.pkl)")
            if ok:
                self.sp.save_project(fname, ds_description={"ds"})

    def export_python(self):
        """Export the project as python file.

        This method is not yet implemented as the functionality is missing in 
        psyplot.
        """
        pass

    def _draw_variable(self, v):
        def func():
            """Visualize variable v"""
            btn = self.variable_buttons[v]
            if not btn.isChecked():
                self.close_sp()
                if self.combo_array.count() > 1:
                    with self.block_widgets(self.combo_array):
                        current = self.combo_array.currentIndex()
                        self.combo_array.setCurrentIndex(current - 1)

            else:
                with self.silence_variable_buttons():
                    for var, btn in self.variable_buttons.items():
                        if var != v:
                            btn.setChecked(False)
                self.make_plot()
            self.refresh()

        return func

    @contextlib.contextmanager
    def silence_variable_buttons(self) -> Iterator[None]:
        """Context manager to disable variable selection buttons."""
        yield self.block_widgets(*self.variable_buttons.values())  # type: ignore

    @property
    def variable(self) -> Union[str, NOTSET_T]:
        """The current variable"""
        for v, btn in self.variable_buttons.items():
            if btn.isChecked():
                return v
        return NOTSET

    @variable.setter
    def variable(self, value: Union[str, NOTSET_T]) -> None:
        with self.silence_variable_buttons():
            for v, btn in self.variable_buttons.items():
                btn.setChecked(v == value)

    @property
    def available_plotmethods(self) -> List[str]:
        """Get the plotmethods that can visualize the selected variable.
        
        Returns
        -------
        list of str
            A list of plotmethod names that can visualize the current
            :attr:`variable`
        """
        v = self.variable
        if v is NOTSET:
            return []
        ret = []
        ds: Dataset = self.ds  # type: ignore
        plot = ds.psy.plot
        for plotmethod in self.plotmethods:
            if plotmethod in plot._plot_methods:
                if getattr(plot, plotmethod).check_data(ds, v, {})[0]:
                    ret.append(plotmethod)
        return ret

    @property
    def plot(self) -> PlotterInterface:
        """Get the plotting function of the currently selected plotmethod."""
        if self.ds is not None:
            return getattr(self.ds.psy.plot, self.plotmethod)
        else:
            raise ValueError(
                "No dataset has yet been selected, so no plot method!")

    @property
    def plot_options(self) -> Dict[str, Any]:
        """Get further keyword arguments for the :attr:`plot` function."""
        if self.ds is not None:
            ret: Dict[str, Any] = self.plotmethod_widget.get_fmts(  # type: ignore
                self.ds.psy[self.variable], True)
            preset = self.preset
            if preset:
                import psyplot.project as psy
                preset = psy.Project.extract_fmts_from_preset(
                    preset, self.plotmethod)
                ret.update(dict(preset))
            return ret
        return {}

    @property
    def plotmethod(self) -> str:
        """Get the name of the current plotmethod."""
        return self.plot_tabs.tabText(self.plot_tabs.currentIndex())

    @plotmethod.setter
    def plotmethod(self, label: str):
        i = next((i for i in range(self.plot_tabs.count())
                  if self.plot_tabs.tabText(i) == label), None)
        if i is not None:
            self.plot_tabs.setCurrentIndex(i)

    @property
    def plotmethods(self) -> List[str]:
        """Get a list of available plotmethods."""
        return list(map(self.plot_tabs.tabText, range(self.plot_tabs.count())))

    @property
    def plotmethod_widget(self) -> PlotMethodWidget:
        """Get widget of the current plotmethod."""
        label = self.plotmethod
        i = next((i for i in range(self.plot_tabs.count())
                  if self.plot_tabs.tabText(i) == label), None)
        return self.plot_tabs.widget(i)

    @property
    def plotmethod_widgets(self) -> Dict[str, PlotMethodWidget]:
        """Get a list of available plotmethod widgets."""
        return dict(zip(self.plotmethods, map(self.plot_tabs.widget,
                                              range(self.plot_tabs.count()))))

    _sp = None

    def get_sp(self) -> Optional[Project]:
        sp = self._sp
        if sp is None:
            return sp
        return self.filter_sp(sp)

    def filter_sp(self, sp: Project, ds_only: bool = False) -> Project:
        """Filter the psyplot project to only include the arrays of :attr:`ds`
        """
        if self._new_plot:
            return None
        if self.ds is None:
            return sp
        num = self.ds.psy.num
        ret = sp[:0]
        for i in range(len(sp)):
            if list(sp[i:i+1].datasets) == [num]:
                ret += sp[i:i+1]
        if ds_only:
            return ret
        arr_name = self.arr_name
        if arr_name is None:
            return ret
        return ret(arr_name=arr_name)

    def new_plot(self) -> None:
        """Select a new variable and make a plot.
        
        This method asks for a variable and them makes a new plot with the
        selected plotmethod.
        
        See Also
        --------
        make_plot: plot the currently selected variable without asking
        """
        if self.ds is not None:
            name, ok = QtWidgets.QInputDialog.getItem(
                self, 'New plot', 'Select a variable',
                self.plotmethod_widget.valid_variables(self.ds))
            if not ok:
                return
            with self.silence_variable_buttons():
                for v, btn in self.variable_buttons.items():
                    btn.setChecked(v == name)
            with self.creating_new_plot():
                self.make_plot()
            self.btn_del.setEnabled(True)
        self.refresh()

    @contextlib.contextmanager
    def creating_new_plot(self) -> Iterator[None]:
        """Tell that we are making a new plot."""
        self._new_plot = True
        yield
        self._new_plot = False

    @property
    def sp(self) -> Optional[Project]:
        """Get the psyplot project of the current plotmethod."""
        return self.plotmethod_widget.sp or None

    @sp.setter
    def sp(self, sp: Optional[Project]):
        if sp is None and (not self._sp or not getattr(
                self._sp, self.plotmethod)):
            pass
        else:
            # first remove the current arrays
            if self.get_sp() and getattr(self.get_sp(), self.plotmethod):
                to_remove = getattr(self.get_sp(), self.plotmethod).arr_names
                _sp: Project = self._sp  # type: ignore
                for i in list(reversed(range(len(_sp)))):
                    if _sp[i].psy.arr_name in to_remove:
                        _sp.pop(i)
            # then add the new arrays
            if sp:
                if self._sp:
                    self._sp.extend(list(sp), new_name=True)
                else:
                    self._sp = sp

    @property
    def data(self) -> Union[psyd.InteractiveList, DataArray]:
        """Get the raw data of the current plot."""
        return self.plotmethod_widget.data

    @property
    def plotter(self) -> Plotter:
        """Get the psyplot plotter of the current plot."""
        return self.plotmethod_widget.plotter

    @property
    def fig(self) -> Figure:
        """Get the figure of the current plot."""
        if self.sp:
            return list(self.sp.figs)[0]

    _animations: Dict[str, FuncAnimation] = {}

    @property
    def animation(self) -> Optional[FuncAnimation]:
        """Get the animation of the current plotmethod."""
        return self._animations.get(self.plotmethod)

    @animation.setter
    def animation(self, ani: Optional[FuncAnimation]):
        if ani is None:
            self._animations.pop(self.plotmethod, None)
        else:
            self._animations[self.plotmethod] = ani

    def make_plot(self) -> None:
        """Make or update the plot of the current variable.

        See Also
        --------
        new_plot: A function to select the variable first
        """
        plotmethods = self.available_plotmethods
        plotmethod = self.plotmethod
        if plotmethod not in plotmethods:
            if not plotmethods:
                QtWidgets.QMessageBox.critical(
                    self, "Visualization impossible",
                    f"Found no plotmethod for variable {self.variable}")
                return
            plotmethod, ok = QtWidgets.QInputDialog.getItem(
                self, "Choose a plot method", "Plot method:", plotmethods)
            if not ok:
                return
            self.plotmethod = plotmethod
        new_v = self.variable
        fmts: Dict[str, Any] = {}
        dims: Dict[Hashable, int] = {}
        if self.sp:
            ds: Dataset = self.ds  # type: ignore
            if not set(self.data.dims) <= set(ds[new_v].dims):
                self.close_sp()
            else:
                for dim in set(ds[new_v].dims) - set(self.data.psy.idims):
                    dims[dim] = 0
                for dim in set(self.data.psy.idims) - set(ds[new_v].dims):
                    del self.data.psy.idims[dim]
        if self.sp:
            if self.data.psy.plotter is None:
                self.data.psy.update(name=self.variable)
                self.data.psy.update(dims=dims, **fmts)
                self.sp.update(replot=True)
            else:
                self.data.psy.update(name=self.variable, dims=dims, **fmts)
            self.show_fig(self.sp)
        else:
            self.ani = None
            self.sp = sp = self.plot(name=self.variable, **self.plot_options)
            self._preset = None
            cid = sp.plotters[0].ax.figure.canvas.mpl_connect(
                'button_press_event', self.display_line)
            self.cids[self.plotmethod] = cid
            self.show_fig(sp)
            descr = sp[0].psy._short_info()
            with self.block_widgets(self.combo_array):
                self.combo_array.addItem(descr)
                self.combo_array.setCurrentText(descr)
        self.expand_current_variable()
        self.enable_navigation()

    @contextlib.contextmanager
    def block_widgets(self, *widgets: QtWidgets.QWidget) -> Iterator[None]:
        """Temporarilly block all signals for the given widgets."""
        for w in widgets:
            w.blockSignals(True)
        yield
        for w in widgets:
            w.blockSignals(False)

    def display_line(self, event: MouseEvent) -> None:
        """Display the line when clicked on a 2D-plot."""
        if not event.inaxes:
            return
        else:
            sl = None
            for widget in map(self.plot_tabs.widget,
                              range(self.plot_tabs.count())):
                if widget.sp and event.inaxes == widget.plotter.ax:
                    sl = widget.get_slice(event.xdata, event.ydata)
                    break
            variable = widget.data.name
            raw_data = widget.data.psy.base.psy[variable]
            if (sl is None or widget.plotmethod not in ['mapplot', 'plot2d'] or
                raw_data.ndim == 2 or
                widget.plotter.ax.figure.canvas.manager.toolbar.mode != ''):
                return
            # check if the mappable contains the event
            if not self.plotter.plot.mappable.contains(event)[0] and (
                    not hasattr(self.plotter.plot, '_wrapped_plot') or
                    not self.plotter.plot._wrapped_plot.contains(event)[0]):
                return
            current_pm = self.plotmethod
            self.plotmethod = 'lineplot'
            linewidget = self.plotmethod_widget
            xdim = linewidget.xdim
            if xdim is None:
                xdim = self.combo_dims.currentText()

            if not linewidget.sp or (linewidget.xdim and
                                     linewidget.xdim not in raw_data.dims):
                with self.silence_variable_buttons():
                    for v, btn in self.variable_buttons.items():
                        btn.setChecked(v == variable)
                self.make_plot()
                linewidget.xdim = xdim
            else:
                linewidget.xdim = xdim
                linewidget.add_line(variable, **sl)
            self.plotmethod = current_pm


    def close_sp(self) -> None:
        """Close the current subproject."""
        sp = self.sp
        self.sp = None
        if sp is not None:
            sp.close(figs=True, data=True, ds=False)

    def show_current_figure(self) -> None:
        """Show figure of the current plotmethod."""
        if self.sp is not None:
            self.show_fig(self.sp)

    def show_fig(self, sp: Project) -> None:
        """Show the first figure in a psyplot project."""
        if len(sp):
            try:
                fig = sp.plotters[0].ax.figure
                fig.canvas.window().show()
                fig.canvas.window().raise_()
            except AttributeError:
                sp.show()

    def switch_tab(self) -> None:
        """Select a valid variable when switching the plotmethod tabs."""
        with self.silence_variable_buttons():
            if self.sp:
                name = self.data.name
            else:
                name = NOTSET
            for v, btn in self.variable_buttons.items():
                btn.setChecked(v == name)
        self.refresh()

    def change_combo_array(self) -> None:
        """Update the iteration dimension depending on the selected variable."""
        with self.block_widgets(self.combo_array):
            sp = self.filter_sp(self._sp, ds_only=True)
            if sp and self.arr_name not in sp.arr_names:
                new_arr = sp.arr_names[0]
                all_arrays = getattr(self._sp, self.plotmethod)
                try:
                    idx_arr = all_arrays.arr_names.index(new_arr)
                except ValueError:
                    idx_arr = 0
                self.combo_array.setCurrentIndex(idx_arr)
                try:
                    vname = self.data.name
                except Exception:
                    vname = self.variable
                if vname is not NOTSET:
                    self.expand_current_variable(vname)
                    self.show_fig(sp[:1])

    def reset_combo_array(self) -> None:
        """Clear and fill the iteration dimension based on :attr:`ds`."""
        curr_arr_name = self.arr_name
        with self.block_widgets(self.combo_array):
            self.combo_array.clear()
            if self._sp:
                all_arrays = getattr(self._sp, self.plotmethod)
                current_ds = self.ds
                if all_arrays:
                    for arr in all_arrays:
                        self.combo_array.addItem(arr.psy._short_info())
                    if curr_arr_name in all_arrays.arr_names:
                        idx_arr = all_arrays.arr_names.index(curr_arr_name)
                        self.combo_array.setCurrentIndex(idx_arr)
                    else:
                        idx_arr = 0
                    self.ds = list(
                        all_arrays[idx_arr:idx_arr+1].datasets.values())[0]
                    if self.ds is not current_ds:
                        with self.block_tree():
                            self.expand_ds_item(self.ds_item)
                            self.expand_current_variable(self.data.name)

    def refresh(self, reset_combo: bool = True) -> None:
        """Refresh the state of this widget.

        This method refreshes the widget based on the selected project, 
        variable, etc.

        Parameters
        ----------
        reset_combo: bool
            If True (default), the :meth:`reset_combo_array` is called
        """
        self.clear_table()

        if reset_combo:
            self.reset_combo_array()

        if self.sp:
            variable = self.data.name
            for action in self._save_preset_actions:
                action.setEnabled(True)
            self.btn_del.setEnabled(True)
            self.btn_export.setEnabled(True)
        else:
            variable = self.variable
            for action in self._save_preset_actions:
                action.setEnabled(False)
            self.btn_del.setEnabled(False)
            self.btn_export.setEnabled(False)


        # refresh variable buttons
        with self.silence_variable_buttons():
            for v, btn in self.variable_buttons.items():
                btn.setChecked(v == variable)

        # refresh tabs
        for i in range(self.plot_tabs.count()):
            w = self.plot_tabs.widget(i)
            w.refresh(self.ds)
        if self.ds is not None and self.variable_buttons:
            valid_variables = self.plotmethod_widget.valid_variables(self.ds)
            for v, btn in self.variable_buttons.items():
                btn.setEnabled(v in valid_variables)
        elif self.ds is None or variable is NOTSET or not self.sp:
            return

        table = self.dimension_table

        if self.sp:
            data = self.data
            ds_data = self.ds[self.variable]

            with self.silence_variable_buttons():
                self.variable_buttons[self.variable].setChecked(True)

            dims: Tuple[Hashable, ...] = ds_data.dims  # type: ignore
            table.setRowCount(ds_data.ndim)
            table.setVerticalHeaderLabels(ds_data.dims)

            # set time, z, x, y info
            for c in 'XYTZ':
                cname = ds_data.psy.get_dim(c)
                if cname and cname in dims:
                    table.setItem(
                        dims.index(cname), 0, QtWidgets.QTableWidgetItem(c))

            for i, dim in enumerate(dims):
                coord = self.ds[dim]
                if 'units' in coord.attrs:
                    table.setItem(
                        i, 4, QtWidgets.QTableWidgetItem(
                            str(coord.attrs['units'])))
                try:
                    coord = list(map("{:1.4f}".format, coord.values))  # type: ignore
                except (ValueError, TypeError):
                    try:
                        coord = coord.to_pandas().dt.to_pydatetime()  # type: ignore
                    except AttributeError:
                        coord = list(map(str, coord.values))  # type: ignore
                    else:
                        coord = [t.isoformat() for t in coord]  # type: ignore
                first = coord[0]
                last = coord[-1]
                table.setItem(
                    i, 1, QtWidgets.QTableWidgetItem(first))
                table.setItem(
                    i, 3, QtWidgets.QTableWidgetItem(last))

                current = data.psy.idims.get(dim)
                if current is not None and isinstance(current, int):
                    table.setCellWidget(
                        i, 2, self.new_dimension_button(dim, coord[current]))

            # update animation checkbox
            dims_to_animate = get_dims_to_iterate(data)

            current_dims_to_animate = list(map(
                self.combo_dims.itemText,
                range(self.combo_dims.count())))
            if dims_to_animate != current_dims_to_animate:
                self.combo_dims.clear()
                self.combo_dims.addItems(dims_to_animate)

        table.resizeColumnsToContents()

        # update plots items
        for ds_item in self.ds_items:
            for item in map(ds_item.child, range(ds_item.childCount())):
                for child in map(item.child, range(item.childCount())):
                    if DatasetTree.is_variable(child):
                        plots_item = ds_item.get_plots_item(child)
                        ds_item.refresh_plots_item(
                            plots_item, child.text(0), self._sp, self.sp)

    def new_dimension_button(
            self, dim: Hashable, label: Any) -> utils.QRightPushButton:
        """Generate a new button to increase of decrease a dimension."""
        btn = utils.QRightPushButton(label)
        btn.clicked.connect(self.increase_dim(str(dim)))
        btn.rightclicked.connect(self.increase_dim(str(dim), -1))
        btn.setToolTip(f"Increase dimension {dim} with left-click, and "
                       "decrease with right-click.")
        return btn

    def update_project(self, *args, **kwargs) -> None:
        """Update the correct project :attr:`sp` and refresh the widget."""
        sp = self.sp
        if sp is not None:
            sp.update(*args, **kwargs)
            self.refresh()

    def increase_dim(self, dim: str, increase: int = 1) -> Callable[[], None]:
        """Get a function to update a specific dimension.

        Parameters
        ----------
        dim: str
            The dimension name
        increase: int
            The number of steps to increase (or decrease) the given `dim`.
        """
        def update():
            i = self.data.psy.idims[dim]
            self.data.psy.update(dims={dim: (i+increase) % self.ds.dims[dim]})
            if self.data.psy.plotter is None:
                self.sp.update(replot=True)
            self.refresh()
        return update


class DatasetWidgetPlugin(DatasetWidget, DockMixin):
    """A :class:`DatasetWidget` plugin for the psyplot GUI.

    This widget can be embeded in the psyplot GUI. Different from the standalone
    :class:`DatasetWidget` class, this one here uses the current psyplot project
    (:func:`psyplot.project.gcp`)
    """

    #: The title of the widget
    title = 'psy-view Dataset viewer'

    #: Display the dock widget at the right side of the GUI
    dock_position = Qt.RightDockWidgetArea

    def __init__(self, *args, **kwargs):
        import psyplot.project as psy
        super().__init__(*args, **kwargs)
        psy.Project.oncpchange.connect(self.oncpchange)

    @property  # type: ignore
    def _sp(self) -> Project:  # type: ignore
        import psyplot.project as psy
        return psy.gcp(True)

    @_sp.setter
    def _sp(self, value):
        pass

    @property
    def sp(self) -> Optional[Project]:
        """Get the psyplot project of the current plotmethod."""
        return self.plotmethod_widget.sp or None

    @sp.setter
    def sp(self, sp: Optional[Project]):
        current = self.get_sp()
        if sp is None or not current:
            return
        elif getattr(current, self.plotmethod, []):

            if len(current) == 1 and len(sp) == 1:
                pass
            # first remove the current arrays
            if current and getattr(current, self.plotmethod):
                to_remove = getattr(self.get_sp(), self.plotmethod).arr_names
                for i in list(reversed(range(len(current)))):
                    if current[i].psy.arr_name in to_remove:
                        current.pop(i)
            # then add the new arrays
            if sp:
                if current:
                    current.extend(list(sp), new_name=True)
                else:
                    current = sp
            current.oncpchange.emit(current)

    @property
    def open_datasets(self) -> Dict[int, Dataset]:
        """Get a mapping from path to dataset number of the open datasets."""
        ret = self._sp.datasets
        ret.update(self._ds_nums)
        return ret

    def close_sp(self) -> None:
        ds = self.ds
        super().close_sp()
        if ds is not None:
            if ds.psy.num not in self._sp.datasets:
                self.set_dataset(ds)

    def oncpchange(self, sp: Optional[Project]) -> None:
        """Update this widget from the current psyplot main (or sub) project."""
        self.reset_combo_array()
        if self.ds is not None and self.ds.psy.num not in self._sp.datasets:
            self.ds = None
            self.disable_navigation()
            self.setup_variable_buttons()
            self.btn_add.setEnabled(False)
            self.btn_del.setEnabled(False)
        elif self.ds is None and self._sp:
            self.set_dataset(next(iter(self._sp.datasets.values())))

    def show_fig(self, sp: Optional[Project]) -> None:
        """Show the figure of the the current subproject."""
        from psyplot_gui.main import mainwindow
        super().show_fig(sp)
        if mainwindow.figures and sp:
            try:
                dock = sp.plotters[0].ax.figure.canvas.manager.window
                dock.widget().show_plugin()
                dock.raise_()
            except AttributeError:
                pass

    def setup_ds_tree(self) -> None:
        """Setup the number of columns and the header of the dataset tree.
        
        Reimplemented to use the :class:`psyplot_gui.content_widget.DatasetTree`
        """
        self.ds_tree = tree = DatasetTree()
        tree.setColumnCount(len(self.ds_attr_columns) + 1)
        tree.setHeaderLabels([''] + self.ds_attr_columns)

    def position_dock(self, main: MainWindow, *args, **kwargs) -> None:
        height = main.help_explorer.dock.size().height()
        main.splitDockWidget(main.help_explorer.dock, self.dock, Qt.Vertical)
        if hasattr(main, 'resizeDocks'):  # qt >= 5.6
            main.resizeDocks([main.help_explorer.dock, self.dock],
                             [height // 2, height // 2], Qt.Vertical)



"""Utility functions for psy-view.

**Disclaimer**

Copyright (C) 2020 Philipp S. Sommer

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.If not, see https: // www.gnu.org / licenses / .
"""
from __future__ import annotations
import os.path as osp

from typing import Callable, Optional, Union, List, cast, TYPE_CHECKING

from PyQt5 import QtWidgets, QtCore, QtGui

if TYPE_CHECKING:
    from PyQt5.QtCore import QEvent  # pylint: disable=no-name-in-module


def get_icon(name: str, ending: str = '.png') -> str:
    return osp.join(osp.dirname(__file__), 'icons', name + ending)


def add_pushbutton(
        label: str,
        connections: Optional[Union[List[Callable], Callable]] = None,
        tooltip: Optional[str] = None,
        layout: Optional[QtWidgets.QLayout] = None,
        icon: bool = False, toolbutton: bool = False, *args, **kwargs
    ) -> Union[QtWidgets.QPushButton, QtWidgets.QToolButton]:
    if icon or toolbutton:
        btn = QtWidgets.QToolButton(*args, **kwargs)
        if icon:
            btn.setIcon(QtGui.QIcon(label))
        else:
            btn.setText(label)
    else:
        btn = QtWidgets.QPushButton(label, *args, **kwargs)
    if tooltip is not None:
        btn.setToolTip(tooltip)
    if connections is not None:
        try:
            iter(connections)  # type: ignore
        except TypeError:
            connections = [connections]  # type: ignore
        connections = cast(List[Callable], connections)
        for con in connections:
            btn.clicked.connect(con)
    if layout is not None:
        layout.addWidget(btn)
    return btn


class QRightPushButton(QtWidgets.QPushButton):
    """A push button that acts differently when right-clicked"""

    rightclicked = QtCore.pyqtSignal()

    def mousePressEvent(self, event: QEvent):
        if event.button() == QtCore.Qt.RightButton:
            self.rightclicked.emit()
            event.accept()
        else:
            return super().mousePressEvent(event)
